"""
This module implements the SYMPROD method.
"""

import itertools

import numpy as np
from sklearn.preprocessing import StandardScaler

from ..base import coalesce
from ..base import OverSampling
from ..base import (NearestNeighborsWithMetricTensor)

from .._logger import logger

__all__ = ['SYMPROD']

class SYMPROD(OverSampling):
    """
    References:
        * Bibtex::

            @article{kunakorntum2020synthetic,
                    title={A Synthetic Minority Based on Probabilistic
                            Distribution (SyMProD) Oversampling for Imbalanced
                            Datasets},
                    author={Kunakorntum, Intouch and Hinthong,
                            Woranich and Phunchongharn, Phond},
                    journal={IEEE Access},
                    volume={8},
                    pages={114692--114704},
                    year={2020},
                    publisher={IEEE}
                }
    """
    categories = [OverSampling.cat_noise_removal,
                  OverSampling.cat_density_based,
                  OverSampling.cat_sample_componentwise,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                    *,
                    proportion=1.0,
                    std_outliers=3,
                    k_neighbors=7,
                    m_neighbors=7,
                    cutoff_threshold=1.25,
                    nn_params=None,
                    n_jobs=1,
                    random_state=None,
                    **_kwargs):
        """
        Constructor of the SYMPROD sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            std_outliers (int): value for removing outliers based on standard
                                deviation of each point
            k_neighbors (int): number of nearest neighbors for calculating distance,
                                closeness factor of each point
            m_neighbors (int): number of nearest neighbors for generating synthetic
                                instances of each point.
            cutoff_threshold (float): threshold for removing minority points where
                                        locating in majority region
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn

        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(std_outliers, "std_outliers", 1)
        self.check_greater_or_equal(k_neighbors, "k_neighbors", 1)
        self.check_greater_or_equal(m_neighbors, "m_neighbors", 1)
        self.check_greater_or_equal(cutoff_threshold, "cutoff_threshold", 0.01)

        self.proportion = proportion
        self.params = {'std_outliers': std_outliers,
                        'k_neighbors': k_neighbors,
                        'm_neighbors': m_neighbors,
                        'cutoff_threshold': cutoff_threshold}
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable paramter combinations.
        Returns:
            list(dict): a list of meaningful paramter combinations
        """
        parameter_combinations = {'proportion': [1.0],
                                  'std_outliers': [3, 4],
                                  'k_neighbors': [5, 7],
                                  'm_neighbors': [5, 7],
                                  'cutoff_threshold': [1.0, 1.25, 1.5]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def noise_removal(self, X, y, scaler):
        """
        Noise removal.

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            scaler (obj): the scaler to be used

        Returns:
            np.array, np.array, np.array, np.array: the minority and majority
                            vectors after noise removal, and all training
                            vectors and labels after noise removal
        """
        def remove_outliers(X):
            return X[np.max(np.abs(scaler.fit_transform(X)), axis=1) \
                                            < self.params['std_outliers']]

        X_min = remove_outliers(X[y == self.min_label])
        X_maj = remove_outliers(X[y == self.maj_label])

        # Apply the original dataset if the dataset is null
        X_min = X[y == self.min_label] if len(X_min) == 0 else X_min
        X_maj = X[y == self.maj_label] if len(X_min) == 0 else X_maj

        # Create X, y after Noise Removal
        # seemingly the purpose of this is to standardize all data together
        y_nr= np.hstack([np.repeat(self.min_label, len(X_min)),
                        np.repeat(self.maj_label, len(X_maj))])
        X_nr= scaler.fit_transform(np.vstack([X_min, X_maj])) # pylint: disable=invalid-name

        # these X_min and X_maj are standardized now
        X_min = X_nr[y_nr == self.min_label]
        X_maj = X_nr[y_nr == self.maj_label]

        return X_min, X_maj, X_nr, y_nr

    def check_all_neighbors_the_same(self, dist, name=''):
        """
        Checks if all neighbors are the same.

        Args:
            dist (np.array): neighbor distances
            name (str): name of the neighborhood
        """
        if np.any(np.nansum(dist, axis=1) == 0.0):
            raise ValueError(f"all {name} samples are the same")

    def normalize(self, array):
        """
        Normalize the array.

        Args:
            array (np.array): the array to normalize

        Returns:
            np.array: the normalized array
        """
        if (array.max() - array.min()) == 0.0:
            array = np.repeat(0.0, len(array))
        else:
            array = (array - array.min()) / (array.max() - array.min())

        return array

    def determine_cf_norms(self, X_maj, X_min, nn_params):
        """
        Determine the closeness factors.

        Args:
            X_maj (np.array): the majority vectors
            X_min (np.array): the minority vectors
            nn_params (dict): the nearest neighbor parameters

        Returns:
            np.array, np.array, np.array, np.array: the minority and majority
                                        closeness factors and the
                                        minority-minority neighborhood
                                        structure and neighbor distances
        """
        nn_maj = NearestNeighborsWithMetricTensor(n_neighbors=len(X_maj),
                                                    n_jobs=self.n_jobs,
                                                    **nn_params).fit(X_maj)
        d_maj_maj, _ = nn_maj.kneighbors(X_maj, return_distance=True)

        nn_min = NearestNeighborsWithMetricTensor(n_neighbors=len(X_min),
                                                    n_jobs=self.n_jobs,
                                                    **nn_params).fit(X_min)
        d_min_min, i_min_min = nn_min.kneighbors(X_min, return_distance=True)

        # Calculate Closeness Factor(CF)
        self.check_all_neighbors_the_same(d_min_min, 'minority')

        cf_min = 1.0 / np.nansum(d_min_min, axis=1)
        cf_norm_min = self.normalize(cf_min)

        self.check_all_neighbors_the_same(d_maj_maj, 'majority')

        cf_maj = 1.0 / np.nansum(d_maj_maj, axis=1)
        cf_norm_maj = self.normalize(cf_maj)

        return cf_norm_min, cf_norm_maj, i_min_min, d_min_min

    def determine_taus(self, X_maj, X_min, nn_params):
        """
        Determine the tau arrays.

        Args:
            X_maj (np.array): the majority vectors
            X_min (np.array): the minority vectors
            nn_params (dict): the nearest neighbor parameters

        Returns:
            np.array, np.array, np.array: the minority and majority taus
                                            and the minority closeness factors
        """
        cf_norm_min, cf_norm_maj, i_min_min, d_min_min = \
            self.determine_cf_norms(X_maj, X_min, nn_params)

        # First, Calculate the distance from minority class to minority class for comparing
        # with majority class

        k_neighbors= np.min([self.params['k_neighbors'], len(X_min) - 1])

        cf_norm_min_idx = np.take(cf_norm_min, i_min_min[:, 1:(k_neighbors + 1)])
        tau_min = np.nanmean(cf_norm_min_idx/(d_min_min[:, 1:(k_neighbors + 1)] + 1),
                             axis=1)

        # Calculate distance from minority to majority
        nn_min_maj = NearestNeighborsWithMetricTensor(n_neighbors=k_neighbors,
                                                        n_jobs=self.n_jobs,
                                                        **nn_params).fit(X_maj)
        d_min_maj, i_min_maj = nn_min_maj.kneighbors(X_min, return_distance=True)

        tau_maj = (np.take(cf_norm_maj, i_min_maj) / (d_min_maj + 1)).mean(axis=1)

        return tau_min, tau_maj, cf_norm_min

    def check_nans(self, array, name=''):
        """
        Checks if NaNs are present in an array.

        Args:
            array (np.array): an array
            name (str): name of the array
        """

        if np.any(np.isnan(array)):
            raise ValueError(f"NaN values in the {name} array")

    def minority_point_selection(self, X_maj, X_min, nn_params):
        """
        Minority point selection.

        Args:
            X_maj (np.array): the majority vectors
            X_min (np.array): the minority vectors
            nn_params (dict): the nearest neighbor parameters

        Returns:
            np.array, np.array, np.array: the probability distribution,
                                        the minority closeness factors and
                                        the selected minority vectors
        """
        tau_min, tau_maj, cf_norm_min = self.determine_taus(X_maj, X_min, nn_params)

        # a cutoff threshold is needed which keeps at least 2 minority samples
        cutoff_threshold = self.params['cutoff_threshold']
        while np.sum(tau_min >= tau_maj * cutoff_threshold) <= 1:
            cutoff_threshold = cutoff_threshold - self.params['cutoff_threshold']/10.0

        logger.info("%s: Cutoff value updated from %f to %f",
                        self.__class__.__name__,
                        self.params['cutoff_threshold'], cutoff_threshold)

        mask = (tau_min >= tau_maj * cutoff_threshold)

        X_min = X_min[mask]
        cf_norm_min = cf_norm_min[mask]

        phi = (tau_min[mask] + 1)/(tau_maj[mask] + 1)
        phi = phi - phi.min()
        prob_dist = phi/phi.sum()

        self.check_nans(prob_dist, 'phi')

        return prob_dist, cf_norm_min, X_min

    def check_enough_neighbors(self, m_neighbors):
        """
        Checks if the number of samples is enough for the m_neighbors parameter

        Args:
            m_neighbors (int): the number of neighbors
        """
        if m_neighbors < 2:
            raise ValueError("Not enough samples for "\
                                f"m_neighbors {self.params['m_neighbors']}")

    def select_neighborhoods(self, X_min, nn_params, n_to_sample):
        """
        Select the neighborhoods used for sampling.

        Args:
            X_min (np.array): the minority samples
            nn_params (dict): the nearest neighbor parameters
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the selected neighbors
        """
        m_neighbors = np.min([self.params['m_neighbors'], len(X_min) - 4])

        self.check_enough_neighbors(m_neighbors)

        nn_min_min = NearestNeighborsWithMetricTensor(m_neighbors + 3,
                                                        n_jobs=self.n_jobs,
                                                        **nn_params).fit(X_min)
        _, i_min_min = nn_min_min.kneighbors(X_min, return_distance=True)

        # it is hardly understandable why M + 3 closest neighbors are selected
        # to take a random sample from
        neighborhoods = np.array(list(itertools.combinations(np.arange(m_neighbors + 3),
                                                    m_neighbors)))

        rand_ind = neighborhoods[self.random_state.choice(len(neighborhoods),
                                                          n_to_sample)]
        nn_selected = np.take(i_min_min, rand_ind)

        return nn_selected

    def instance_synthesis(self,
                            *,
                            prob_dist,
                            n_to_sample,
                            X_min,
                            nn_params,
                            cf_norm_min):
        """
        Instance synthesis.

        Args:
            prob_dist (np.array): the probability distribution
            n_to_sample (int): the number of samples to generate
            X_min (np.array): the minority vectors
            nn_params (dict): the nearest neighbor parameters
            cf_norm_min (np.array): the minority closeness factors

        Returns:
            np.array: the generated samples
        """
        min_reference = self.random_state.choice(len(prob_dist),
                                                    n_to_sample,
                                                    p=prob_dist,
                                                    replace=True)

        nn_selected = self.select_neighborhoods(X_min, nn_params, n_to_sample)

        synthetic_points = np.zeros(shape=(n_to_sample, X_min[0].shape[0]))

        feature_value = np.concatenate([X_min[nn_selected],
                                        X_min[min_reference][:, None, :]],
                                        axis=1)
        CF = np.hstack([cf_norm_min[nn_selected], # pylint: disable=invalid-name
                        cf_norm_min[min_reference][:, None]])
        dirichlet_param = np.ones(self.params['m_neighbors'] + 1) \
                                    * (self.params['m_neighbors'] + 2)
        random_beta = self.random_state.dirichlet(dirichlet_param,
                                                  size=min_reference.shape[0])

        calculate_CF = random_beta * CF # pylint: disable=invalid-name
        calculate_CF = (calculate_CF.T / np.sum(calculate_CF, axis=1)).T # pylint: disable=invalid-name

        feature_value = feature_value * calculate_CF[:, :, None]

        synthetic_points = np.sum(feature_value, axis=1)

        return synthetic_points

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        ####################
        # I. Noise Removal #
        ####################

        scaler = StandardScaler()

        X_min, X_maj, X_nr, y_nr = self.noise_removal(X, y, scaler) # pylint: disable=invalid-name

        ################################
        # II. Minority point selection #
        ################################

        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
                self.metric_tensor_from_nn_params(nn_params, X_nr, y_nr)

        try:
            prob_dist, cf_norm_min, X_min = \
                self.minority_point_selection(X_maj, X_min, nn_params)
        except ValueError as valueerror:
            return self.return_copies(X, y, valueerror.args[0])

        ###########################
        # III. Instance Synthesis #
        ###########################

        try:
            synthetic_points = self.instance_synthesis(prob_dist=prob_dist,
                                                    n_to_sample=n_to_sample,
                                                    X_min=X_min,
                                                    nn_params=nn_params,
                                                    cf_norm_min=cf_norm_min)
        except ValueError as valueerror:
            return self.return_copies(X, y, valueerror.args[0])

        return (np.vstack([X, np.vstack(scaler.inverse_transform(synthetic_points))]),
                np.hstack([y, np.repeat(self.min_label, len(synthetic_points))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'std_outliers': self.params['std_outliers'],
                'k_neighbors': self.params['k_neighbors'],
                'm_neighbors': self.params['m_neighbors'],
                'cutoff_threshold': self.params['cutoff_threshold'],
                'nn_params': self.nn_params,
                **OverSampling.get_params(self)}



"""
This module implements the SUNDO method.
"""

import numpy as np

from sklearn.preprocessing import MinMaxScaler

from ..base import coalesce
from ..base import (NearestNeighborsWithMetricTensor,
                                pairwise_distances_mahalanobis)
from ..base import OverSampling

from .._logger import logger
_logger= logger

__all__= ['SUNDO']

class SUNDO(OverSampling):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{sundo,
                            author={Cateni, S. and Colla, V. and Vannucci, M.},
                            booktitle={2011 11th International Conference on
                                        Intelligent Systems Design and
                                        Applications},
                            title={Novel resampling method for the
                                    classification of imbalanced datasets for
                                    industrial and other real-world problems},
                            year={2011},
                            volume={},
                            number={},
                            pages={402-407},
                            keywords={decision trees;pattern classification;
                                        sampling methods;support vector
                                        machines;resampling method;imbalanced
                                        dataset classification;industrial
                                        problem;real world problem;
                                        oversampling technique;undersampling
                                        technique;support vector machine;
                                        decision tree;binary classification;
                                        synthetic dataset;public dataset;
                                        industrial dataset;Support vector
                                        machines;Training;Accuracy;Databases;
                                        Intelligent systems;Breast cancer;
                                        Decision trees;oversampling;
                                        undersampling;imbalanced dataset},
                            doi={10.1109/ISDA.2011.6121689},
                            ISSN={2164-7151},
                            month={Nov}}
    """

    categories = [OverSampling.cat_changes_majority,
                  OverSampling.cat_application,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 *,
                 nn_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)

        self.check_n_jobs(n_jobs, 'n_jobs')

        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        _ = raw # pylint hack
        return [{}]

    def generate_samples(self, X_min, X_maj, nn_params, N): # pylint: disable=invalid-name
        """
        Generate samples.

        Args:
            X_min (np.array): the minority vectors
            X_maj (np.array): the majority vectors
            nn_params (dict): the nearest neighbor parameters
            N (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """
        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=1,
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nnmt.fit(X_maj)

        stds = np.std(X_min, axis=0)
        # At one point the algorithm says to keep those points which are
        # the most distant from majority samples, and not leaving any minority
        # sample isolated. This can be implemented by generating multiple
        # samples for each point and keep the one most distant from the
        # majority samples.
        base_indices = self.random_state.choice(np.arange(X_min.shape[0]), N)
        base_vectors = X_min[base_indices]

        n_trials = 3
        offsets = self.random_state.normal(size=(base_vectors.shape[0],
                                                 n_trials,
                                                 base_vectors.shape[1]))
        samples = base_vectors[:, None, :] + offsets * stds[None, None, :]
        sample_stack = samples.reshape(base_vectors.shape[0] * n_trials, base_vectors.shape[1])
        dist, _ = nnmt.kneighbors(sample_stack)
        dist = dist[:, 0].reshape(base_vectors.shape[0], n_trials)
        samples = samples[np.arange(samples.shape[0]), np.argmax(dist, axis=1)]

        return samples

    def determine_removal(self, X_maj, nn_params, N): # pylint: disable=invalid-name
        """
        Determine which majority samples need to be removed.

        Args:
            X_maj (np.array): the majority vectors
            nn_params (dict): the nearest neighbor parameters
            N (int): the number of samples to generate

        Returns:
            np.array: the indices to remove
        """
        # normalize
        mms = MinMaxScaler()
        X_maj_normalized = mms.fit_transform(X_maj) # pylint: disable=invalid-name

        # computing the distance matrix
        distm = pairwise_distances_mahalanobis(X_maj_normalized,
                                             tensor=nn_params.get('metric_tensor', None))

        # len(X_maj) offsets for the diagonal 0 elements, 2N because
        # every distances appears twice
        threshold = sorted(distm.flatten())[np.min([X_maj.shape[0] + 2 * N,
                                                 distm.shape[0]**2 - 1])]
        np.fill_diagonal(distm, np.inf)

        # extracting the coordinates of pairs closer than threshold
        pairs_to_break = np.where(distm < threshold)
        pairs_to_break = np.vstack(pairs_to_break)

        # sorting the pairs, otherwise both points would be removed
        pairs_to_break.sort(axis=1)

        # uniqueing the coordinates - the final number might be less than N
        to_remove = np.unique(pairs_to_break[0])

        return to_remove

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        X_min = X[y == self.min_label]
        X_maj = X[y == self.maj_label]

        n_1 = len(X_min)
        n_0 = len(X) - n_1
        N = int(np.rint(0.5*n_0 - 0.5*n_1 + 0.5)) # pylint: disable=invalid-name

        if N == 0:
            return self.return_copies(X, y, "N is 0")

        # generating minority samples
        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
                    self.metric_tensor_from_nn_params(nn_params, X, y)

        samples = self.generate_samples(X_min, X_maj, nn_params, N)

        # Extending the minority dataset with the new samples
        X_min_extended = np.vstack([X_min, samples]) # pylint: disable=invalid-name

        # Removing N elements from the majority dataset
        to_remove = self.determine_removal(X_maj, nn_params, N)

        # removing the selected elements
        X_maj_cleaned = np.delete(X_maj, to_remove, axis=0) # pylint: disable=invalid-name

        return (np.vstack([X_min_extended, X_maj_cleaned]),
                np.hstack([np.repeat(self.min_label, len(X_min_extended)),
                           np.repeat(self.maj_label, len(X_maj_cleaned))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the parallelizable Evaluation Job
"""

import os
import hashlib
import time
import warnings

import numpy as np

from ..base import instantiate_obj, load_dict, dump_dict, check_if_damaged
from ..base import calculate_all_scores

__all__ = ['EvaluationJob']

class EvaluationJob:
    """
    An evaluation job to be executed in parallel
    """

    def __init__(self,
                oversampling,
                classifiers,
                *,
                cache_path=None,
                serialization="json",
                reset=False):
        """
        Constructor of the object

        oversampling (str/dict): an oversampling or a filename containing an
                                    oversampling
        classifiers (list(tuple)): classifier specification
        cache_path (str): the cache path
        serialization (str): 'json'/'pickle'
        reset (str): whether to recompute the oversampling
        """
        self.oversampling = oversampling
        self.classifiers = classifiers
        self.cache_path = cache_path
        self.serialization = serialization
        self.reset = reset

        if self.cache_path is not None:
            os.makedirs(self.cache_path, exist_ok=True)

    def get_params(self, deep=False):
        """
        Returns the parameters of the object.

        Args:
            deep (bool): deep parameters

        Returns:
            dict: the dictionary of parameters
        """
        _ = deep # pylint workaround
        return {'oversampling': self.oversampling,
                'classifiers': self.classifiers,
                'cache_path': self.cache_path,
                'serialization': self.serialization,
                'reset': self.reset}

    def target_filename_damaged(self, classifier, oversampling):
        """
        Determine the target filename and whether it is damaged

        Args:
            classifier (tuple): a classifier specification
            oversampling (dict): an oversampling

        Returns:
            str, bool: the target filename and whether it is damaged
        """
        target_filename = None

        label0 = str(classifier).encode('utf-8')
        label1 = str((oversampling['oversampler'])).encode('utf-8')

        hashcode = str(hashlib.md5(label0 + label1).hexdigest())

        damaged = False
        if self.cache_path is not None:
            repeat_idx = oversampling['fold_descriptor']['repeat_idx']
            filename = f"evaluation_{repeat_idx:04}"\
                        f"_{oversampling['fold_descriptor']['fold_idx']:04}"\
                        f"_{hashcode}.{self.serialization}"
            target_filename = os.path.join(self.cache_path, filename)

            damaged = check_if_damaged(target_filename, self.serialization)

        return target_filename, damaged

    def do_evaluation(self):
        """
        do the evaluation

        Returns:
            dict/str: an evaluation dictionary or a filename
        """
        oversampling = self.oversampling

        all_results = []

        if isinstance(oversampling, str):
            oversampling = load_dict(oversampling, oversampling.split('.')[-1],
                                ['X_train', 'X_test', 'y_train', 'y_test'])

        for classifier in self.classifiers:
            target_filename, damaged = self.target_filename_damaged(classifier,
                                                                    oversampling)

            if self.cache_path is not None and not damaged and not self.reset:
                all_results.append(target_filename)
                continue

            classifier_obj = instantiate_obj(classifier)

            error = None
            warning_list = None

            begin_timestamp = time.time()

            with warnings.catch_warnings(record=True) as warning:
                try:
                    y_pred = classifier_obj.fit(oversampling['X_train'],
                                                oversampling['y_train'])\
                                            .predict_proba(oversampling['X_test'])
                except ValueError as value_error:
                    error = str(value_error)
                    y_pred = np.zeros(shape=(oversampling['y_test'].shape[0], 2))
                except RuntimeError as runtime_error:
                    error = str(runtime_error)
                    y_pred = np.zeros(shape=(oversampling['y_test'].shape[0], 2))

                warning_list = [(str(warn.category), warn.message) for warn in warning]

            evaluation = {'fold_descriptor': oversampling['fold_descriptor'],
                    'oversampler': oversampling['oversampler'],
                    'runtime': time.time() - begin_timestamp,
                    'y_pred': y_pred[:, 1],
                    'y_test': oversampling['y_test'],
                    'scores': calculate_all_scores(oversampling['y_test'],
                                                y_pred[:, 1]),
                    'classifier': classifier[1],
                    'classifier_params': classifier[2],
                    'classifier_module': classifier[0],
                    'error': error,
                    'warning': str(warning_list)}

            if self.cache_path is not None:
                dump_dict(evaluation, target_filename, self.serialization,
                        ['y_pred', 'y_test'])
                all_results.append(target_filename)
                continue

            all_results.append(evaluation)

        return all_results



"""
This module brings together all functionalities.
"""

from ._version import __version__

__author__ = "György Kovács"
__license__ = "MIT"
__email__ = "gyuriofkovacs@gmail.com"

from .oversampling import *
from .queries import *
from .multiclassoversampling import *

from . import config
from . import base
from . import classifiers
from . import datasets
from . import evaluation
from . import multiclassoversampling
from . import queries
from . import visualization



"""
This module implements the GASMOTE method.
"""

import numpy as np

from sklearn.model_selection import KFold
from sklearn.tree import DecisionTreeClassifier

from ..base import coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['GASMOTE']

class GASMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @Article{gasmote,
                        author="Jiang, Kun
                        and Lu, Jing
                        and Xia, Kuiliang",
                        title="A Novel Algorithm for Imbalance Data
                                Classification Based on Genetic
                                Algorithm Improved SMOTE",
                        journal="Arabian Journal for Science and
                                    Engineering",
                        year="2016",
                        month="Aug",
                        day="01",
                        volume="41",
                        number="8",
                        pages="3255--3266",
                        issn="2191-4281",
                        doi="10.1007/s13369-016-2179-2",
                        url="https://doi.org/10.1007/s13369-016-2179-2"
                        }
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_memetic,
                  OverSamplingSimplex.cat_sample_ordinary,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 maxn=7,
                 n_pop=10,
                 popl3=5,
                 pm=0.3,
                 pr=0.2,
                 Ge=10,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            n_neighbors (int): number of neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            maxn (int): maximum number of samples to generate per minority
                        instances
            n_pop (int): size of population
            popl3 (int): number of crossovers
            pm (float): mutation probability
            pr (float): selection probability
            Ge (int): number of generations
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        nn_params = coalesce(nn_params, {})

        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater_or_equal(maxn, "maxn", 1)
        self.check_greater_or_equal(n_pop, "n_pop", 1)
        self.check_in_range(pm, "pm", [0, 1])
        self.check_in_range(pr, "pr", [0, 1])
        self.check_greater_or_equal(Ge, "Ge", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.n_neighbors = n_neighbors
        self.nn_params = nn_params
        self.ga_params = {'maxn': maxn,
                            'n_pop': n_pop,
                            'popl3': popl3,
                            'pm': pm,
                            'pr': pr,
                            'Ge': Ge}
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        return cls.generate_parameter_combinations({'n_neighbors': [7],
                                                    'maxn': [2, 3, 4],
                                                    'n_pop': [10],
                                                    'popl3': [4],
                                                    'pm': [0.3],
                                                    'pr': [0.2],
                                                    'Ge': [10]}, raw)

    def samples_from_conf(self, X_min, ind, conf):
        """
        Generate samples from a configuration.

        Args:
            X_min (np.array): the minority samples
            ind (np.array): the neighborhoods
            conf (np.array): a configuration

        Returns:
            np.array: the generated samples
        """
        samples = [np.zeros(shape=(0, X_min.shape[1]), dtype=float)]
        for idx, conf_i in enumerate(conf):
            samples.append(self.sample_simplex(X=X_min[[idx]],
                                                indices=ind[[idx]],
                                                n_to_sample=conf_i,
                                                X_vertices=X_min))
        return np.vstack(samples)

    def calculate_score(self, tests, preds):
        """
        Calculate the objective function score

        Args:
            tests (np.array): the test labels
            preds (np.array): the predicted labels

        Returns:
            float: the score
        """
        # compute fitness measure
        t_p = np.sum(np.logical_and(tests == self.min_label, tests == preds))
        t_n = np.sum(np.logical_and(tests == self.maj_label, tests == preds))
        f_p = np.sum(np.logical_and(tests == self.maj_label, tests != preds))
        f_n = np.sum(np.logical_and(tests == self.min_label, tests != preds))
        sens = t_p/(t_p + f_n)
        spec = t_n/(f_p + t_n)

        return np.sqrt(sens*spec)

    def fitness(self, *, conf, X_min, ind, X, y, kfold):
        """
        Evluate fitness of configuration

        Args:
            conf (list(list)): configuration
            X_min (np.array): the minority samples
            ind (np.array): the neighborhood relation
            X (np.array): all vectors
            y (np.array): all target labels
            kfold (obj): cross validation object

        Returns:
            float: the fitness score
        """
        # generate new samples
        samples = self.samples_from_conf(X_min, ind, conf)

        #samples = []
        #for idx, conf_i in enumerate(conf):
        #    for _ in range(conf_i):
        #        X_b = X_min[self.random_state.choice(ind[idx][1:])]
        #        samples.append(self.sample_between_points(X_min[idx], X_b))

        X_new = np.vstack([X, samples])
        y_new = np.hstack([y, np.repeat(self.min_label, len(samples))])

        # execute kfold cross validation
        preds, tests = [], []
        for train, test in kfold.split(X_new):
            clas = DecisionTreeClassifier(random_state=self._random_state_init)
            clas.fit(X_new[train], y_new[train])
            preds.append(clas.predict(X_new[test]))
            tests.append(y_new[test])
        preds = np.hstack(preds)
        tests = np.hstack(tests)

        return self.calculate_score(tests, preds)

    def crossover(self, conf_a, conf_b):
        """
        Crossover

        Args:
            conf_a (list): configuration to crossover
            conf_b (list): configuration to crossover

        Returns:
            list, list: the configurations after crossover
        """
        for _ in range(self.ga_params['popl3']):
            kdx = self.random_state.randint(len(conf_a))
            conf_a = np.hstack([conf_a[:kdx], conf_b[kdx:]])
            conf_b = np.hstack([conf_b[:kdx], conf_a[kdx:]])
        return conf_a, conf_b

    def mutation(self, conf, ge_):
        """
        Mutation

        Args:
            conf (list): configuration to mutate
            ge (int): iteration number

        Returns:
            list: the mutated configuration
        """
        conf = conf.copy()
        if self.random_state.random_sample() >= self.ga_params['pm']:
            for idx, conf_i in enumerate(conf):
                rand = self.random_state.random_sample()
                rand = rand**((1 - ge_/self.ga_params['Ge'])**3)
                if self.random_state.randint(2) == 0:
                    tmp = (self.ga_params['maxn'] - conf_i) * rand
                    conf[idx] = int(conf_i + tmp)
                else:
                    conf[idx] = int(conf_i - (conf_i - 0)*rand)
        return conf

    # generate initial population
    def init_pop(self, X_min):
        """
        Initialize the population

        Args:
            X_min (np.array): the minority vectors

        Returns:
            np.array: an initial population
        """
        return self.random_state.randint(self.ga_params['maxn'], size=X_min.shape[0])

    def genetic_algorithm(self, *, X_min, ind, X, y, kfold):
        """
        The genetic algorithm

        Args:
            X_min (np.array): the minority vectors
            ind (np.array): the neighborhood relations
            X (np.array): all feature vectors
            y (np.array): all target labels
            kfold (obj): a k-fold cross-validation object
        """
        population = [[self.init_pop(X_min), 0] for _ in range(self.ga_params['n_pop'])]

        # calculate fitness values
        for pop in population:
            pop[1] = self.fitness(conf=pop[0],
                                    X_min=X_min,
                                    ind=ind,
                                    X=X,
                                    y=y,
                                    kfold=kfold)

        # start iteration
        ge_ = 0
        while ge_ < self.ga_params['Ge']:
            # sorting population in descending order by fitness scores
            population = sorted(population, key=lambda x: -x[1])

            # selection operation (Step 2)
            n_keep = int(self.ga_params['n_pop']*self.ga_params['pr'])
            population_new = []
            population_new.extend(population[:n_keep])
            population_new.extend(population[:(self.ga_params['n_pop'] - n_keep)])
            population = population_new

            # crossover
            for _ in range(int(self.ga_params['n_pop']/2)):
                pop_0 = population[self.random_state.randint(self.ga_params['n_pop'])][0]
                pop_1 = population[self.random_state.randint(self.ga_params['n_pop'])][0]
                conf_a, conf_b = self.crossover(pop_0, pop_1)
                population.append([conf_a, self.fitness(conf=conf_a,
                                                        X_min=X_min,
                                                        ind=ind,
                                                        X=X,
                                                        y=y,
                                                        kfold=kfold)])
                population.append([conf_b, self.fitness(conf=conf_b,
                                                        X_min=X_min,
                                                        ind=ind,
                                                        X=X,
                                                        y=y,
                                                        kfold=kfold)])

            # mutation
            for _ in range(int(self.ga_params['n_pop']/2)):
                pop_0 = population[self.random_state.randint(self.ga_params['n_pop'])][0]
                conf_a = self.mutation(pop_0, ge_)
                population.append([conf_a, self.fitness(conf=conf_a,
                                                        X_min=X_min,
                                                        ind=ind,
                                                        X=X,
                                                        y=y,
                                                        kfold=kfold)])

            ge_ = ge_ + 1

        # sorting final population
        population = sorted(population, key=lambda x: -x[1])

        # return best configuration
        return population[0][0]

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        X_min = X[y == self.min_label]

        # fitting nearest neighbors model to find minority neighbors of
        #  minority samples
        n_neighbors = min([self.n_neighbors + 1, len(X_min)])

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                        n_jobs=self.n_jobs,
                                                        **(nn_params))
        nnmt.fit(X_min)
        ind = nnmt.kneighbors(X_min, return_distance=False)
        kfold = KFold(min([len(X), 5]))

        conf = self.genetic_algorithm(X_min=X_min,
                                        ind=ind,
                                        X=X,
                                        y=y,
                                        kfold=kfold)

        # generate final samples
        samples = self.samples_from_conf(X_min, ind, conf)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'maxn': self.ga_params['maxn'],
                'n_pop': self.ga_params['n_pop'],
                'popl3': self.ga_params['popl3'],
                'pm': self.ga_params['pm'],
                'pr': self.ga_params['pr'],
                'Ge': self.ga_params['Ge'],
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
Testing the Safe_Level_SMOTE.
"""
import numpy as np

from smote_variants import Safe_Level_SMOTE

def test_specific():
    """
    Oversampler specific testing
    """
    obj = Safe_Level_SMOTE()

    X = np.array([[1, 2], [2, 3], [3, 4], [4, 5],
                    [5, 6], [6, 7], [8, 9]])
    y = np.array([0, 1, 0, 1, 0, 1, 0])

    assert obj.sample(X, y)[0].shape[0] > 0

    X, y = obj.generate_samples(np.array([]), np.array([]), np.array([]), 2)
    assert len(X) == 0



"""
This module implements the Borderline_SMOTE methods.
"""

import numpy as np

from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from ..base import mode, coalesce_dict, coalesce

from .._logger import logger
_logger= logger

__all__= ['Borderline_SMOTE1',
          'Borderline_SMOTE2']

def determine_danger_remove_noise(*, X, y, X_min,
                                    nn_params,
                                    n_neighbors,
                                    n_jobs,
                                    maj_label):
    """
    Determine in-danger items and remove noise.

    Args:
        X (np.array): all samples
        y (np.array): all target values
        X_min (np.array): all minority samples
        nn_params (dict): the nearest neighbors parameters
        n_neighbors (int): number of neighbors
        n_jobs (int): number of jobs
        maj_label (int): majority label

    Returns:
        np.array, np.array: X_min, X_danger
    """
    n_neighbors = np.min([len(X), n_neighbors + 1])

    nnmt= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                            n_jobs=n_jobs,
                                            **(nn_params))
    nnmt.fit(X)
    indices= nnmt.kneighbors(X_min, return_distance=False)

    # determining minority samples in danger
    noise = []
    danger = []
    for idx, row in enumerate(indices):
        if (n_neighbors - 1) == sum(y[row[1:]] == maj_label):
            noise.append(idx)
        elif mode(y[row[1:]]) == maj_label:
            danger.append(idx)

    X_danger = X_min[danger] # pylint: disable=invalid-name
    X_min = np.delete(X_min, np.array(noise).astype(int), axis=0)

    return X_min, X_danger

class Borderline_SMOTE1(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @InProceedings{borderlineSMOTE,
                            author="Han, Hui
                            and Wang, Wen-Yuan
                            and Mao, Bing-Huan",
                            editor="Huang, De-Shuang
                            and Zhang, Xiao-Ping
                            and Huang, Guang-Bin",
                            title="Borderline-SMOTE: A New Over-Sampling Method
                                     in Imbalanced Data Sets Learning",
                            booktitle="Advances in Intelligent Computing",
                            year="2005",
                            publisher="Springer Berlin Heidelberg",
                            address="Berlin, Heidelberg",
                            pages="878--887",
                            isbn="978-3-540-31902-3"
                            }
    """

    categories = [OverSamplingSimplex.cat_sample_ordinary,
                  OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_borderline,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 k_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                    to sample e.g. 1.0 means that after
                                    sampling the number of minority samples
                                    will be equal to the number of majority
                                    samples
            n_neighbors (int): control parameter of the nearest neighbor
                                    technique for determining the borderline
            k_neighbors (int): control parameter of the nearest neighbor
                                    technique for sampling
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)

        self.check_greater_or_equal(proportion, 'proportion', 0)
        self.check_greater_or_equal(n_neighbors, 'n_neighbors', 1)
        self.check_greater_or_equal(k_neighbors, 'k_neighbors', 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.k_neighbors = k_neighbors
        self.nn_params= coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'k_neighbors': [3, 5, 7]}

        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        # determining number of samples to be generated
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        # fitting model
        X_min = X[y == self.min_label]

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params,
                                                                        X, y)

        X_min, X_danger = determine_danger_remove_noise(X=X, y=y, # pylint: disable=invalid-name
                                                        X_min=X_min,
                                                        nn_params=nn_params,
                                                        n_neighbors=self.n_neighbors,
                                                        n_jobs=self.n_jobs,
                                                        maj_label=self.maj_label)

        if len(X_min) < 2 or len(X_danger) == 0:
            return self.return_copies(X, y, "The number of samples "\
                    f"after preprocessing X_min ({len(X_min)}) "\
                    f" X_danger ({len(X_danger)} is not enough for sampling.")

        # fitting nearest neighbors model to minority samples
        k_neigh = min([len(X_min), self.k_neighbors + 1])

        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=k_neigh,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_min)
        indices= nnmt.kneighbors(X_danger, return_distance=False)

        # X_danger items are also present in X_min
        indices = indices[:, 1:]

        samples = self.sample_simplex(X=X_danger,
                                        indices=indices,
                                        n_to_sample=n_to_sample,
                                        X_vertices=X_min)

        # original implementation
        # generating samples near points in danger
        # base_indices = self.random_state.choice(list(range(len(X_danger))),
        #                                        n_to_sample)
        #distribution = np.repeat(1, k_neigh-1)
        #distribution = distribution / np.sum(distribution)
        #
        #neighbor_indices = self.random_state.choice(list(range(1, k_neigh)),
        #                                            n_to_sample,
        #                                            p=distribution)
        #
        #X_base = X_danger[base_indices]
        #X_neighbor = X_min[indices[base_indices, neighbor_indices]]
        #
        #samples = X_base + \
        #    np.multiply(self.random_state.rand(
        #        n_to_sample, 1), X_neighbor - X_base)

        return (np.vstack([X, samples]),
                np.hstack([y, np.hstack([self.min_label]*n_to_sample)]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'k_neighbors': self.k_neighbors,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}


class Borderline_SMOTE2(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @InProceedings{borderlineSMOTE,
                            author="Han, Hui
                            and Wang, Wen-Yuan
                            and Mao, Bing-Huan",
                            editor="Huang, De-Shuang
                            and Zhang, Xiao-Ping
                            and Huang, Guang-Bin",
                            title="Borderline-SMOTE: A New Over-Sampling
                                    Method in Imbalanced Data Sets Learning",
                            booktitle="Advances in Intelligent Computing",
                            year="2005",
                            publisher="Springer Berlin Heidelberg",
                            address="Berlin, Heidelberg",
                            pages="878--887",
                            isbn="978-3-540-31902-3"
                            }
    """

    categories = [OverSamplingSimplex.cat_sample_ordinary,
                  OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_borderline,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 k_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and
                                n_min to sample e.g. 1.0 means that after
                                sampling the number of minority samples
                                will be equal to the number of majority
                                samples
            n_neighbors (int): control parameter of the nearest neighbor
                                technique for determining the borderline
            k_neighbors (int): control parameter of the nearest neighbor
                                technique for sampling
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            preferential (bool): whether to use preferential connectivity in the
                                    neighborhoods
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)

        self.check_greater_or_equal(proportion, 'proportion', 0)
        self.check_greater_or_equal(n_neighbors, 'n_neighbors', 1)
        self.check_greater_or_equal(k_neighbors, 'k_neighbors', 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.k_neighbors = k_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'k_neighbors': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        # determining number of samples to be generated
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        # fitting nearest neighbors model
        X_min = X[y == self.min_label]

        X_min, X_danger = determine_danger_remove_noise(X=X, y=y, # pylint: disable=invalid-name
                                                        X_min=X_min,
                                                        nn_params=nn_params,
                                                        n_neighbors=self.n_neighbors,
                                                        n_jobs=self.n_jobs,
                                                        maj_label=self.maj_label)

        if len(X_min) < 2 or len(X_danger) == 0:
            return self.return_copies(X, y, "The number of samples "\
                    f"after preprocessing X_min ({len(X_min)}) "\
                    f" X_danger ({len(X_danger)} is not enough for sampling.")

        # fitting nearest neighbors model to samples
        k_neigh = self.k_neighbors + 1
        k_neigh = min([k_neigh, len(X)])

        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=k_neigh,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X)
        indices= nnmt.kneighbors(X_danger, return_distance=False)

        # X_danger items are also present in X
        indices = indices[:, 1:]

        vertex_weights = np.repeat(1.0, len(X))
        vertex_weights[y == self.maj_label] = 0.5

        samples = self.sample_simplex(X=X_danger,
                                        indices=indices,
                                        n_to_sample=n_to_sample,
                                        X_vertices=X,
                                        vertex_weights=vertex_weights)

        # generating the samples
        #base_indices = self.random_state.choice(
        #    list(range(len(X_danger))), n_to_sample)
        #
        #distribution = np.repeat(1, k_neigh-1)
        #distribution = distribution / np.sum(distribution)
        #
        #neighbor_indices = self.random_state.choice(
        #    list(range(1, k_neigh)), n_to_sample, p=distribution)
        #
        #X_base = X_danger[base_indices]
        #X_neighbor = X[indices[base_indices, neighbor_indices]]
        #diff = X_neighbor - X_base
        #rand = self.random_state.rand(n_to_sample, 1)
        #mask = y[neighbor_indices] == self.maj_label
        #rand[mask] = rand[mask]*0.5
        #
        #samples = X_base + np.multiply(rand, diff)

        return (np.vstack([X, samples]),
                np.hstack([y, np.hstack([self.min_label]*n_to_sample)]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'k_neighbors': self.k_neighbors,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the SMOTE_D method.
"""

import numpy as np

from ..base import fix_density
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling
from .._logger import logger
_logger= logger

__all__= ['SMOTE_D']

class SMOTE_D(OverSampling):
    """
    References:
        * BibTex::

            @InProceedings{smote_d,
                            author="Torres, Fredy Rodr{\'i}guez
                            and Carrasco-Ochoa, Jes{\'u}s A.
                            and Mart{\'i}nez-Trinidad, Jos{\'e} Fco.",
                            editor="Mart{\'i}nez-Trinidad, Jos{\'e} Francisco
                            and Carrasco-Ochoa, Jes{\'u}s Ariel
                            and Ayala Ramirez, Victor
                            and Olvera-L{\'o}pez, Jos{\'e} Arturo
                            and Jiang, Xiaoyi",
                            title="SMOTE-D a Deterministic Version of SMOTE",
                            booktitle="Pattern Recognition",
                            year="2016",
                            publisher="Springer International Publishing",
                            address="Cham",
                            pages="177--188",
                            isbn="978-3-319-39393-3"
                            }

    Notes:
        * Copying happens if two points are the neighbors of each other.
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 k=3,
                 *,
                 nn_params={},
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            k (int): number of neighbors in nearest neighbors component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)

        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(k, "k", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.k = k # pylint: disable=invalid-name
        self.nn_params = nn_params
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'k': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def generate_samples(self, X_min, ind, counts_ij):
        """
        Generate samples.

        Args:
            X_min (np.array): minority sampes
            ind (np.array): neighborhood structure
            counts_ij (np.array): the counts (density)

        Returns:
            np.array: the generated samples
        """
        # do the sampling
        samples = [np.zeros(shape=(0, X_min.shape[1]))]
        for idx, _ in enumerate(X_min):
            for jdx in range(ind.shape[1]-1):
                while counts_ij[idx][jdx] > 0:
                    if self.random_state.random_sample() < counts_ij[idx][jdx]:
                        translation = X_min[ind[idx][jdx + 1]] - X_min[idx]
                        weight = counts_ij[idx][jdx] + 1
                        samples.append(X_min[idx] + translation/weight)
                    counts_ij[idx][jdx] = counts_ij[idx][jdx] - 1

        return np.vstack(samples)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
                self.metric_tensor_from_nn_params(nn_params, X, y)

        # fitting nearest neighbors model
        n_neighbors = min([len(X_min), self.k+1])

        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_min)
        dist, ind = nnmt.kneighbors(X_min)

        # extracting standard deviations of distances
        stds = np.std(dist[:, 1:], axis=1)

        # estimating sampling density
        p_i = fix_density(stds)

        # the other component of sampling density
        p_ij = np.array([fix_density(dist[idx, 1:])
                            for idx in range(dist.shape[0])])
        #p_ij = dist[:, 1:]/np.sum(dist[:, 1:], axis=1)[:, None]

        # number of samples to generate between minority points
        counts_ij = n_to_sample * p_i[:, None] * p_ij

        samples = self.generate_samples(X_min, ind, counts_ij)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'k': self.k,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the SMOTE_PSOBAT method.
"""

import numpy as np

from sklearn.model_selection import KFold
from sklearn.tree import DecisionTreeClassifier

from ..base import coalesce, coalesce_dict
from ..base import OverSampling
from ._smote import SMOTE

from .._logger import logger
_logger= logger

__all__= ['SMOTE_PSOBAT']

class SMOTE_PSOBAT(OverSampling):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{smote_psobat,
                            author={Li, J. and Fong, S. and Zhuang, Y.},
                            booktitle={2015 3rd International Symposium on
                                        Computational and Business
                                        Intelligence (ISCBI)},
                            title={Optimizing SMOTE by Metaheuristics with
                                    Neural Network and Decision Tree},
                            year={2015},
                            volume={},
                            number={},
                            pages={26-32},
                            keywords={data mining;particle swarm
                                        optimisation;pattern classification;
                                        data mining;classifier;metaherustics;
                                        SMOTE parameters;performance
                                        indicators;selection optimization;
                                        PSO;particle swarm optimization
                                        algorithm;BAT;bat-inspired algorithm;
                                        metaheuristic optimization algorithms;
                                        nearest neighbors;imbalanced dataset
                                        problem;synthetic minority
                                        over-sampling technique;decision tree;
                                        neural network;Classification
                                        algorithms;Neural networks;Decision
                                        trees;Training;Optimization;Particle
                                        swarm optimization;Data mining;SMOTE;
                                        Swarm Intelligence;parameter
                                        selection optimization},
                            doi={10.1109/ISCBI.2015.12},
                            ISSN={},
                            month={Dec}}

    Notes:
        * The parameters of the memetic algorithms are not specified.
        * I have checked multiple paper describing the BAT algorithm, but the
            meaning of "Generate a new solution by flying randomly" is still
            unclear.
        * It is also unclear if best solutions are recorded for each bat, or
            the entire population.
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_uses_clustering,
                  OverSampling.cat_sample_ordinary,
                  OverSampling.cat_memetic,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 *,
                 maxit=50,
                 c1=0.3,
                 c2=0.1,
                 c3=0.1,
                 alpha=0.9,
                 gamma=0.9,
                 method='bat',
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            maxit (int): maximum number of iterations
            c1 (float): intertia weight of PSO
            c2 (float): attraction of local maximums in PSO
            c3 (float): attraction of global maximum in PSO
            alpha (float): alpha parameter of the method
            gamma (float): gamma parameter of the method
            method (str): optimization technique to be used
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}

        super().__init__(random_state=random_state)
        self.check_greater_or_equal(maxit, "maxit", 1)
        self.check_greater_or_equal(c1, "c1", 0)
        self.check_greater_or_equal(c2, "c2", 0)
        self.check_greater_or_equal(c3, "c3", 0)
        self.check_greater_or_equal(alpha, "alpha", 0)
        self.check_greater_or_equal(gamma, "gamma", 0)
        self.check_isin(method, "method", ['pso', 'bat'])
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.params = {'maxit': maxit,
                        'c1': c1,
                        'c2': c2,
                        'c3': c3,
                        'alpha': alpha,
                        'gamma': gamma,
                        'n_pop': 10,
                        'n_keep': 5,
                        'f_max': 10}
        self.method = method
        self.nn_params = coalesce(nn_params, {})
        self.ss_params = coalesce_dict(ss_params, ss_params_default)
        self.n_jobs = n_jobs

        # a reasonable range of proportions
        self.proportion_range = np.array([0.1, 2.0])
        self.k_range = np.array([1, 10])

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        bat_pc = cls.generate_parameter_combinations({'maxit': [50],
                                                      'alpha': [0.7, 0.9],
                                                      'gamma': [0.7, 0.9],
                                                      'method': ['bat']}, raw)
        pso_pc = cls.generate_parameter_combinations({'maxit': [50],
                                                      'c1': [0.2, 0.5],
                                                      'c2': [0.1, 0.2],
                                                      'c3': [0.1, 0.2],
                                                      'method': ['pso']}, raw)
        if not raw:
            bat_pc.extend(pso_pc)
            return bat_pc

        return {'maxit': [50],
                'alpha': [0.7, 0.9],
                'gamma': [0.7, 0.9],
                'method': ['bat', 'pso'],
                'c1': [0.2, 0.5],
                'c2': [0.1, 0.2],
                'c3': [0.1, 0.2]}

    def pred_test_vectors(self, *, proportion, K, nn_params, X, y):
        """
        Determine the prediction and test vectors.

        Args:
            proportion (float): the proportion to sample
            K (int): the number of neighbors
            nn_params (dict): the nearest neighbor parameters
            X (np.array): all training vectors
            y (np.array): all target labels

        Returns:
            np.array, np.array: the predictions and the test labels
        """
        ss_params = self.ss_params.copy()
        ss_params['n_dim'] = np.min([ss_params['n_dim'], int(K)])

        X_samp, y_samp = SMOTE(proportion=proportion,
                            n_neighbors=K,
                            nn_params=nn_params,
                            ss_params=ss_params,
                            n_jobs=self.n_jobs,
                            random_state=self._random_state_init).sample(X, y)

        # doing k-fold cross validation
        kfold = KFold(5)
        preds = []
        tests = []
        for train, test in kfold.split(X_samp):
            dtree = DecisionTreeClassifier(random_state=self._random_state_init)
            dtree.fit(X_samp[train], y_samp[train])
            preds.append(dtree.predict(X_samp[test]))
            tests.append(y_samp[test])
        preds = np.hstack(preds)
        tests = np.hstack(tests)

        return preds, tests

    def calculate_scores(self, truep, truen, falsep, falsen):
        """
        Calculate the target scores.

        Args:
            truep (int): the number of true positives
            truen (int): the number of true negatives
            falsep (int): the number of false positives
            falsen (int): the number of false negatives

        Returns:
            float, float: kappa, accuracy
        """
        p_o = (truep + truen) / (truep + falsen + truen + falsep)
        term_0 = (truep + falsen) * (truep + falsep) \
                / (truep + falsen + truen + falsep)**2
        term_1 = (falsep + truen)*(falsen + truen) \
                / (truep + falsen + truen + falsep)**2
        p_e =  term_0 + term_1

        kappa = (p_o - p_e)/(1.0 - p_e)

        return kappa, p_o

    def evaluate(self, *, K, proportion, nn_params, X, y):
        """
        Evaluate given configuration

        Args:
            K (int): number of neighbors in nearest neighbors component
            proportion (float): proportion of missing data to generate
            nn_params (dict): nearest neighbor parameters
            X (np.array): the training vectors
            y (np.array): the target labels

        Returns:
            float, float: kappa and accuracy scores
        """
        preds, tests = self.pred_test_vectors(proportion=proportion,
                                                K=K,
                                                nn_params=nn_params,
                                                X=X,
                                                y=y)

        # computing the kappa score
        truep = np.sum(np.logical_and(preds == tests,
                                    tests == self.min_label))
        falsen = np.sum(np.logical_and(preds != tests,
                                    tests == self.min_label))
        truen = np.sum(np.logical_and(preds == tests,
                                    tests == self.maj_label))
        falsep = np.sum(np.logical_and(preds != tests,
                                    tests == self.maj_label))

        return self.calculate_scores(truep, truen, falsep, falsen)

    def init_particle_or_bat(self):
        """
        Initialize a PSO particle.

        Args:
            k_range (tuple): the range of k values
            proportion_range (tuple): the range of proportion values

        Returns:
            np.array: an initial particle
        """
        k_rand = self.random_state.randint(self.k_range[0], self.k_range[1])
        rand = self.random_state.random_sample()
        diff = self.proportion_range[1] - self.proportion_range[0]
        vect = rand * diff + self.proportion_range[0]
        return np.array([k_rand, vect])

    def init_population(self):
        """
        Initialize the population.

        Returns:
            np.array: the initial population
        """
        return np.array([self.init_particle_or_bat() \
                            for _ in range(self.params['n_pop'])])

    def pso_update_local_scores(self,
                                scores,
                                particles,
                                local_scores,
                                local_best):
        """
        Update the local scores according to the rule in the paper.

        Args:
            scores (np.array): the new scores
            particles (np.array): the particles
            local_scores (np.array): the local best scores
            local_best (np.array): the local best particles

        Returns:
            np.array, np.array, bool: local scores, local best particles,
                                        not changed flag
        """
        not_changed = True

        mask = ((np.min(np.vstack([local_scores[:, 0],
                                  scores[:, 0]]).T, axis=1) > 0.4)
                                & (local_scores[:, 1] > scores[:, 1]))

        local_scores[mask] = scores[mask].copy()
        local_best[mask] = particles[mask].copy()
        if np.any(mask):
            not_changed = False

        mask = (~mask) & (scores[:, 0] > 0.4) & (local_scores[:, 0] <= 0.4)
        local_scores[mask] = scores[mask].copy()
        local_best[mask] = particles[mask].copy()

        if np.any(mask):
            not_changed = False

        return local_scores, local_best, not_changed

    def pso_update_global_scores(self,
                                    scores,
                                    particles,
                                    global_scores,
                                    global_best):
        """
        Update global scores for PSO

        Args:
            scores (np.array): the actual scores
            particles (np.array): the particles
            global_scores (np.array): the global scores
            global_best (np.array): the globally best particle

        Returns:
            np.array, np.array, bool: the global best score, best particle,
                                        the not changed flag
        """
        not_changed = True
        mask = ((np.min(np.vstack([np.repeat(global_scores[0], len(scores)),
                                                scores[:, 0]]).T, axis=1) > 0.4)
                    & (global_scores[1] > scores[:, 1]))
        if np.any(mask):
            max_idx = np.argmax(scores[mask, 1])
            global_scores = scores[mask][max_idx].copy()
            global_best = particles[mask][max_idx].copy()
            not_changed = False

        mask = (~mask) & (scores[:, 0] > 0.4) & (global_scores[0] <= 0.4)
        if np.any(mask):
            max_idx = np.argmax(scores[mask, 1])
            global_scores = scores[mask][max_idx].copy()
            global_best = particles[mask][max_idx].copy()
            not_changed = False

        return (global_scores, global_best, not_changed)

    def pso_update_velocities(self, particles, velocities, local_best, global_best):
        """
        Update the PSO velocities

        Args:
            particles (np.array): the particles
            velocities (np.array): the actual velocities
            local_best (np.array): the locally best particles
            global_best (np.array): the globally best particle

        Returns:
            np.array: the updated velocities
        """
        # update velocities
        velocities = self.params['c1'] * velocities \
                        + (local_best - particles) * self.params['c2'] \
                        + (global_best - particles) * self.params['c3']

        ratios = np.abs(velocities[:, 0]) / (self.k_range[1] - self.k_range[0])
        velocities[ratios > 1.0, 0] = \
                velocities[ratios > 1.0, 0] / np.ceil(ratios[ratios > 1.0])

        proportion_diff = self.proportion_range[1] - self.proportion_range[0]
        ratios = np.abs(velocities[:, 1]) / proportion_diff
        velocities[ratios > 1.0, 1] = \
                velocities[ratios > 1.0, 1] / np.ceil(ratios[ratios > 1.0])

        return velocities

    def pso_vectorized(self, X, y, nn_params):
        """
        PSO optimization

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            nn_params (dict): the nearest neighbor parameters

        Returns:
            np.array: the best K and proportion values
        """

        # initial particles

        particles = self.init_population()
        # initial velocities
        velocities = np.array([np.array([0, 0])
                                        for _ in range(self.params['n_pop'])])

        local_best = np.array([particles[i].copy()
                                        for i in range(self.params['n_pop'])])
        local_scores = np.zeros((self.params['n_pop'], 2))
        global_best = particles[0].copy()
        global_scores = np.array([0.0, 0.0])

        # executing the particle swarm optimization
        not_changed = 0
        for _ in range(self.params['maxit']):
            # if the configurations didn't change for 10 iterations, stop
            if not_changed > len(particles):
                break

            not_changed = not_changed + 1

            scores = np.array([self.evaluate(K=int(particle[0]),
                                        proportion=particle[1],
                                        nn_params=nn_params,
                                        X=X, y=y) for particle in particles])

            (local_scores, local_best, changed) = \
                self.pso_update_local_scores(scores,
                                                particles,
                                                local_scores,
                                                local_best)
            (global_scores, global_best, changed) = \
                self.pso_update_global_scores(scores,
                                                particles,
                                                global_scores,
                                                global_best)

            not_changed = not_changed * changed

            velocities = self.pso_update_velocities(particles,
                                                    velocities,
                                                    local_best,
                                                    global_best)

            # update positions
            particles = particles + velocities
            particles[:, 0] = np.clip(particles[:, 0],
                                        self.k_range[0],
                                        self.k_range[1])
            particles[:, 1] = np.clip(particles[:, 1],
                                        self.proportion_range[0],
                                        self.proportion_range[1])

        return global_best

    def bat_check_global(self, scores, global_scores):
        """
        Check the global performances for the bat optimization

        Args:
            scores (np.array): the actual scores
            global_scores (np.array): the globally best scores

        Returns:
            int: the index maximizing the global score or -1 if no improvement
                 happened
        """
        max_idx = -1
        tmp0 = np.repeat(global_scores[0], len(scores))
        tmp1 = scores[:, 0]

        mask = ((np.min(np.vstack([tmp0, tmp1]).T, axis=1) > 0.4)
                    & (global_scores[1] > scores[:, 1]))
        if np.any(mask):
            max_idx = np.argmax(scores[mask, 1])

        mask = (~mask) & (scores[:, 0] > 0.4) & (global_scores[0] <= 0.4)
        if np.any(mask):
            max_idx = np.argmax(scores[mask, 1])

        return max_idx

    def bat_check_local(self, scores, local_scores):
        """
        Check the performance locally.

        Args:
            scores (np.array): the actual scores
            local_scores (np.array): the local scores

        Returns:
            np.array: the mask of particles that improved
        """
        local_scores = np.array([local[0] for local in local_scores])
        mask = ((np.min(np.vstack([local_scores[:, 0], scores[:, 0]]).T, axis=1) > 0.4)
                                & (local_scores[:, 1] > scores[:, 1]))

        mask = (~mask) & (scores[:, 0] > 0.4) & (local_scores[:, 0] <= 0.4)

        return mask

    def bat_local_search(self, bats, local_best, local_scores, pulse_loudness):
        """
        The local search algorithm.

        Args:
            bats (np.array): the bats
            local_best (np.array): the locally best bats
            local_scores (np.array): the local scores
            pulse_loudness (np.array): the pulse and loudness parameters

        Returns:
            np.array: the updated bats
        """
        n_pop = self.params['n_pop']

        local_lens = np.sum(local_scores[:, :, 0] >= 0.0, axis=1)
        mask = self.random_state.random_sample(n_pop) > pulse_loudness[:, 1]

        random_best_idx = np.floor(local_lens \
                * self.random_state.random_sample(n_pop)).astype(int)
        random_best = local_best[np.arange(n_pop), random_best_idx][mask]
        rand = (self.random_state.random_sample(size=random_best.shape) - 0.5) * 2.0
        rand = rand / 20.0
        rand[:, 0] = rand[:, 0] * (self.k_range[1] - self.k_range[0])
        rand[:, 1] = rand[:, 1] * (self.proportion_range[1] - self.proportion_range[0])
        bats[mask] = random_best + rand * np.mean(pulse_loudness[:, 2])

        bats[:, 0] = np.clip(bats[:, 0],
                                self.k_range[0],
                                self.k_range[1])
        bats[:, 1] = np.clip(bats[:, 1],
                                self.proportion_range[0],
                                self.proportion_range[1])

        return bats

    def bat_update_local(self, *, bats, scores, local_best, local_scores, improved_local):
        """
        Update the local scores for the bat optimization

        Args:
            bats (np.array): the bats
            scores (np.array): the actual scores
            local_best (np.array): the locally best bats
            local_scores (np.array): the local scores
            improved_local (np.array): the mask of particles that improved locally

        Returns:
            np.array, np.array: the updated local scores, and the locally best bats
        """
        local_lens = np.sum(local_scores[:, :, 0] >= 0.0, axis=1)
        local_lens[improved_local] = np.minimum(local_lens[improved_local] + 1,
                                                self.params['n_keep'])
        local_best[improved_local, local_lens[improved_local]] = \
                                                        bats[improved_local]
        local_scores[improved_local, local_lens[improved_local]] = \
                                                        scores[improved_local]

        sorting = local_scores[:,:,0].argsort(axis=1)[:,::-1]
        sorting = sorting + \
                        (np.arange(self.params['n_pop']) * self.params['n_keep'])[:, None]
        n_total = self.params['n_pop']*self.params['n_keep']

        local_scores = local_scores.reshape(n_total, 2)[sorting]
        local_scores = local_scores.reshape(self.params['n_pop'],
                                            self.params['n_keep'],
                                            2)

        local_best = local_best.reshape(n_total, 2)[sorting]
        local_best = local_best.reshape(self.params['n_pop'],
                                        self.params['n_keep'],
                                        2)

        return local_scores, local_best

    def bat_update_pulse_loudness(self,
                                    pulse_loudness,
                                    improved_local,
                                    iteration):
        """
        Update the pulse and loudness values.

        Args:
            pulse_loudness (np.array): the pulse and loudness values
            improved_local (np.array): the mask of bats that improved locally
            iteration (int): the iteration number

        Returns:
            np.array: the updated pulse-loudness parameters
        """
        random = self.random_state.random_sample(self.params['n_pop'])
        mask = random[improved_local] < pulse_loudness[improved_local, 1]
        pulse_loudness[improved_local, 2][mask] = \
                pulse_loudness[improved_local, 2][mask] * self.params['alpha']
        pulse_loudness[improved_local, 1][mask] = \
                pulse_loudness[improved_local, 1][mask] \
                    * (1 - np.exp(-self.params['gamma'] * iteration))

        return pulse_loudness

    def bat_update_local_scores(self,
                                    *,
                                    bats,
                                    scores,
                                    local_best,
                                    local_scores,
                                    pulse_loudness,
                                    iteration):
        """
        Update the local scores in the bat algorithm.

        Args:
            bats (np.array): all bats
            scores (np.array): the actual scores
            local_best (np.array): the locally best bats
            local_scores (np.array): the local scores
            pulse_loudness (np.array): the pulse-loudness parameters
            iteration (int): the iteration number

        Returns:
            np.array, np.array, np.array, bool: the updated locally best bats,
                                                local scores, pulse-loudness
                                                values and the flag indicating
                                                any update
        """
        improved_local = self.bat_check_local(scores, local_scores)
        local_scores, local_best = self.bat_update_local(bats=bats,
                                                        scores=scores,
                                                        local_best=local_best,
                                                        local_scores=local_scores,
                                                        improved_local=improved_local)
        pulse_loudness = self.bat_update_pulse_loudness(pulse_loudness,
                                                        improved_local,
                                                        iteration)

        return local_best, local_scores, pulse_loudness, np.any(improved_local)

    def bat_update_global_scores(self,
                                    *,
                                    bats,
                                    scores,
                                    global_best,
                                    global_scores,
                                    pulse_loudness):
        """
        Update the global scores in the bat algorithm.

        Args:
            bats (np.array): the bats
            scores (np.array): the actual scores
            global_best (np.array): the globally best bat
            global_scores (np.array): the globally best scores
            pulse_loudness (np.array): the pulse-loudness parameters

        Returns:
            np.array, np.array, bool: the updated globally best bat, globally
                                        best scores and the improvement flag
        """
        improved_global = self.bat_check_global(scores, global_scores)

        rand = self.random_state.random_sample()

        if (improved_global >= 0 and rand < pulse_loudness[improved_global, 2]):
            global_best = bats[improved_global]
            global_scores = scores[improved_global]

        return global_best, global_scores, improved_global >= 0

    def bat_vectorized(self, X, y, nn_params):
        """
        BAT optimization

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            nn_params (dict): the nearest neighbor parameters

        Returns:
            np.array: the best K and proportion values
        """

        # initial bat positions
        bats = self.init_population()
        # initial velocities
        velocities = np.zeros(shape=(self.params['n_pop'], 2))

        # best configurations of particles
        local_best = np.array([[bats[i].copy()]*self.params['n_keep']
                                for i in range(len(bats))])

        local_scores = np.zeros(shape=(self.params['n_pop'], self.params['n_keep'], 2))
        local_scores[:, :, :] = -1

        # scores of best configurations of particles
        global_best = bats[0].copy()
        global_scores = np.array([0.0, 0.0])

        # columns: pulse frequency, pulse rate, loudness
        pulse_loudness = self.random_state.random_sample(size=(self.params['n_pop'], 3))

        not_changed = 0
        for iteration in range(self.params['maxit']):
            not_changed = not_changed + 1

            if not_changed > 10:
                break

            # update frequencies
            pulse_loudness[:, 0] = \
                self.random_state.random_sample(size=self.params['n_pop']) \
                                                            * self.params['f_max']

            velocities = velocities + (bats - global_best) \
                                            * pulse_loudness[:, 0, None]

            bats = bats + velocities

            bats = self.bat_local_search(bats,
                                            local_best,
                                            local_scores,
                                            pulse_loudness)

            scores = np.array([self.evaluate(K=int(bat[0]),
                                             proportion=bat[1],
                                             nn_params=nn_params,
                                             X=X, y=y) for bat in bats])

            (local_best, local_scores, pulse_loudness, changed) = \
                self.bat_update_local_scores(bats=bats,
                                                scores=scores,
                                                local_best=local_best,
                                                local_scores=local_scores,
                                                pulse_loudness=pulse_loudness,
                                                iteration=iteration)

            global_best, global_scores, changed = \
                    self.bat_update_global_scores(bats=bats,
                                                    scores=scores,
                                                    global_best=global_best,
                                                    global_scores=global_scores,
                                                    pulse_loudness=pulse_loudness)

            not_changed = not_changed * changed

        return global_best

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
                self.metric_tensor_from_nn_params(nn_params, X, y)

        # a reasonable range of nearest neighbors to use with SMOTE
        self.k_range = np.array([1, np.min([np.sum(y == self.min_label),
                                            self.k_range[1]])])

        if self.method == 'pso':
            best_combination = self.pso_vectorized(X, y, nn_params)
        elif self.method == 'bat':
            best_combination = self.bat_vectorized(X, y, nn_params)

        ss_params = self.ss_params.copy()
        ss_params['n_dim'] = np.min([ss_params['n_dim'], int(best_combination[0])])

        return SMOTE(proportion=best_combination[1],
                     n_neighbors=int(best_combination[0]),
                     nn_params=nn_params,
                     ss_params=ss_params,
                     n_jobs=self.n_jobs,
                     random_state=self._random_state_init).sample(X, y)

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'maxit': self.params['maxit'],
                'c1': self.params['c1'],
                'c2': self.params['c2'],
                'c3': self.params['c3'],
                'alpha': self.params['alpha'],
                'gamma': self.params['gamma'],
                'method': self.method,
                'nn_params': self.nn_params,
                'ss_params': self.ss_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the parallelizable Sampling Job
"""

import os
import hashlib
import time
import warnings

from ..base import instantiate_obj, load_dict, dump_dict, check_if_damaged

__all__= ['SamplingJob']

class SamplingJob:
    """
    An oversampling job to be executed in parallel
    """
    def __init__(self,
                 folding,
                 oversampler,
                 oversampler_params,
                 *,
                 scaler=('sklearn.preprocessing', 'StandardScaler', {}),
                 cache_path=None,
                 serialization="json",
                 reset=False):
        """
        Constructor of the object

        folding (str/dict): a folding or a filename containing a folding
        oversampler (str): name of the oversampler class
        oversampler_params (dict): the dictionary of oversampler params
        scaler (obj): a scaler object
        cache_path (str): the cache path
        serialization (str): 'json'/'pickle'
        reset (str): whether to recompute the oversampling
        """
        self.folding = folding
        self.oversampler = oversampler
        self.oversampler_params = oversampler_params
        self.scaler = scaler
        self.cache_path = cache_path
        self.serialization = serialization
        self.reset = reset

        if self.cache_path is not None:
            os.makedirs(self.cache_path, exist_ok=True)

    def get_params(self, deep=False):
        """
        Returns the parameters of the object.

        Args:
            deep (bool): deep parameters

        Returns:
            dict: the dictionary of parameters
        """
        _ = deep # pylint workaround
        return {'folding': self.folding,
                'oversampler': self.oversampler,
                'oversampler_params': self.oversampler_params,
                'cache_path': self.cache_path,
                'serialization': self.serialization,
                'reset': self.reset}

    def target_filename_damaged(self, folding):
        """
        Determine the target filename and whether it is damaged.

        Args:
            folding (obj): a folding

        Returns:
            str, bool: the target filename and the damaged flag
        """
        target_filename = None

        label = str((self.oversampler, self.oversampler_params)).encode('utf-8')
        hashcode = str(hashlib.md5(label).hexdigest())

        damaged = False
        if self.cache_path is not None:
            filename = f"oversampling_{folding['fold_descriptor']['repeat_idx']:04}"\
                        f"_{folding['fold_descriptor']['fold_idx']:04}"\
                        f"_{hashcode}.{self.serialization}"
            target_filename = os.path.join(self.cache_path, filename)

            damaged = check_if_damaged(target_filename, self.serialization)

        return target_filename, damaged

    def do_oversampling(self):
        """
        do the oversampling

        Returns:
            dict/str: an oversampling dictionary or a filename
        """
        folding = self.folding

        if isinstance(folding, str):
            folding = load_dict(folding, folding.split('.')[-1],
                                ['X_train', 'y_train', 'X_test', 'y_test'])

        target_filename, damaged = self.target_filename_damaged(folding)

        if self.cache_path is not None and not damaged and not self.reset:
            return target_filename

        oversampler_obj = instantiate_obj(('smote_variants',
                                            self.oversampler,
                                            self.oversampler_params))

        X_train = folding['X_train']
        X_test = folding['X_test']

        if self.scaler is not None:
            scaler = instantiate_obj(self.scaler)
            X_train = scaler.fit_transform(X_train)
            X_test = scaler.transform(X_test)

        oversampling = {'fold_descriptor': folding['fold_descriptor'],
                        'y_test': folding['y_test'],
                        'oversampler': oversampler_obj.get_params()}

        begin_timestamp = time.time()
        with warnings.catch_warnings(record=True) as warning:
            try:
                X_samp, y_samp = oversampler_obj.sample(folding['X_train'],
                                                        folding['y_train'])
                X_test = oversampler_obj.preprocessing_transform(folding['X_test'])
                oversampling = {**oversampling,
                                'X_train': X_samp,
                                'y_train': y_samp,
                                'X_test': X_test,
                                'runtime': time.time() - begin_timestamp,
                                'error': None,
                                'warning': None}
            except ValueError as value_error:
                oversampling = {**oversampling,
                                'X_train': folding['X_train'],
                                'y_train': folding['y_train'],
                                'X_test': folding['X_test'],
                                'runtime': time.time() - begin_timestamp,
                                'error': str(value_error)}
            except RuntimeError as runtime_error:
                oversampling = {**oversampling,
                                'X_train': folding['X_train'],
                                'y_train': folding['y_train'],
                                'X_test': folding['X_test'],
                                'runtime': time.time() - begin_timestamp,
                                'error': str(runtime_error)}

            oversampling = {**oversampling,
                            'warning': str([(str(warn.category), warn.message)
                                                            for warn in warning])}

        if self.cache_path is not None:
            dump_dict(oversampling, target_filename, self.serialization,
                        ['X_train', 'y_train', 'X_test', 'y_test'])
            return target_filename

        return oversampling



"""
This module implements the ISOMAP_Hybrid method.
"""
import warnings

import numpy as np

from sklearn.manifold import Isomap

from ..config import suppress_external_warnings
from ..base import coalesce, coalesce_dict
from ..base import OverSampling
from ._smote import SMOTE
from ..noise_removal import NeighborhoodCleaningRule

from .._logger import logger
_logger= logger

__all__= ['ISOMAP_Hybrid']

class ISOMAP_Hybrid(OverSampling):
    """
    References:
        * BibTex::

            @inproceedings{isomap_hybrid,
                             author = {Gu, Qiong and Cai, Zhihua and Zhu, Li},
                             title = {Classification of Imbalanced Data Sets by
                                        Using the Hybrid Re-sampling Algorithm
                                        Based on Isomap},
                             booktitle = {Proceedings of the 4th International
                                            Symposium on Advances in
                                            Computation and Intelligence},
                             series = {ISICA '09},
                             year = {2009},
                             isbn = {978-3-642-04842-5},
                             location = {Huangshi, China},
                             pages = {287--296},
                             numpages = {10},
                             doi = {10.1007/978-3-642-04843-2_31},
                             acmid = {1691478},
                             publisher = {Springer-Verlag},
                             address = {Berlin, Heidelberg},
                             keywords = {Imbalanced data set, Isomap, NCR,
                                            Smote, re-sampling},
                            }
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_noise_removal,
                  OverSampling.cat_dim_reduction,
                  OverSampling.cat_changes_majority,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_components=3,
                 smote_n_neighbors=5,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_components (int): number of components
            smote_n_neighbors (int): number of neighbors in SMOTE sampling
            n_jobs (int): number of parallel jobs
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}

        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater_or_equal(n_components, "n_components", 1)
        self.check_greater_or_equal(smote_n_neighbors, "smote_n_neighbors", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.isomap_params = {}
        self.nn_params = coalesce(nn_params, {})
        self.ss_params = coalesce_dict(ss_params, ss_params_default)
        self.smote_n_neighbors = smote_n_neighbors
        self.n_jobs = n_jobs

        self.isomap = Isomap(n_neighbors=n_neighbors,
                             n_components=n_components,
                             n_jobs=n_jobs)

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'n_components': [2, 3, 4],
                                  'smote_n_neighbors': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        with warnings.catch_warnings():
            if suppress_external_warnings():
                warnings.simplefilter("ignore")
            X_trans = self.isomap.fit_transform(X, y) # pylint: disable=invalid-name

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= \
                    self.metric_tensor_from_nn_params(nn_params, X_trans, y)

        X_sm, y_sm = SMOTE(proportion=self.proportion, # pylint: disable=invalid-name
                           n_neighbors=self.smote_n_neighbors,
                           nn_params=nn_params,
                           ss_params=self.ss_params,
                           n_jobs=self.n_jobs,
                           random_state=self._random_state_init).sample(X_trans, y)

        ncr = NeighborhoodCleaningRule(n_jobs=self.n_jobs,
                                      nn_params=nn_params)
        X_final, y_final= ncr.remove_noise(X_sm, y_sm)

        if np.sum(y_final) <= 2 or np.sum(y_final == 0) <= 2:
            return self.return_copies(X_sm, y_sm, "too many samples removed "\
                                    "returning isomapped dataset")

        return X_final, y_final

    def preprocessing_transform(self, X):
        """
        Transforms new data by the trained isomap

        Args:
            X (np.array): new data

        Returns:
            np.array: the transformed data
        """
        return self.isomap.transform(X)

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.isomap.n_neighbors,
                'nn_params': self.nn_params,
                'ss_params': self.ss_params,
                'n_components': self.isomap.n_components,
                'smote_n_neighbors': self.smote_n_neighbors,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module tests the NoSMOTE object
"""

from smote_variants import NoSMOTE
from smote_variants.datasets import load_illustration_2_class

def test_nosmote():
    """
    Testing the NoSMOTE technique
    """

    obj = NoSMOTE()

    assert len(NoSMOTE.parameter_combinations()) == 1

    dataset = load_illustration_2_class()

    assert len(obj.sampling_algorithm(dataset['data'],
                                  dataset['target'])[0]) > 0



"""
This module implements flexible oversampler queries.
"""

import inspect

import numpy as np

from ..base import OverSampling, SimplexSamplingMixin
from ..noise_removal import NoiseFilter
from ..oversampling import (A_SUWO, ADASYN, ADG, ADOMS, AHC, AMSCO,
AND_SMOTE, ANS, ASMOBD, Assembled_SMOTE, Borderline_SMOTE1, Borderline_SMOTE2,
CBSO, CCR, CE_SMOTE, cluster_SMOTE, CURE_SMOTE, DBSMOTE, DE_oversampling, DEAGO,
distance_SMOTE, DSMOTE, DSRBF, E_SMOTE, Edge_Det_SMOTE, G_SMOTE, GASMOTE,
Gaussian_SMOTE, polynom_fit_SMOTE_star, polynom_fit_SMOTE_bus, polynom_fit_SMOTE_poly,
polynom_fit_SMOTE_mesh, Gazzah, IPADE_ID, ISMOTE, ISOMAP_Hybrid, KernelADASYN,
kmeans_SMOTE, Lee, LLE_SMOTE, LN_SMOTE, LVQ_SMOTE, MCT, MDO, MOT2LD, MSYN,
MWMOTE, NDO_sampling, NEATER, NRAS, NRSBoundary_SMOTE, NT_SMOTE,
OUPS, PDFOS, ProWSyn, Random_SMOTE, ROSE, RWO_sampling, Safe_Level_SMOTE,
SDSMOTE, Selected_SMOTE, SL_graph_SMOTE, SMMO, SMOBD, SMOTE_Cosine,
SMOTE_D, SMOTE_ENN, SMOTE_FRST_2T, SMOTE_IPF, SMOTE_OUT, SMOTE_PSO,
SMOTE_PSOBAT, SMOTE_RSB, SMOTE_TomekLinks, SMOTE, SN_SMOTE, SOI_CJ,
SOMO, SPY, SSO, Stefanowski, SUNDO, Supervised_SMOTE, SVM_balance, SYMPROD,
TRIM_SMOTE, V_SYNTH, VIS_RST, MSMOTE)

from ._runtimes import runtimes

__all__= ['get_all_oversamplers',
          'get_metric_learning_oversamplers',
          'get_simplex_sampling_oversamplers',
          'get_multiclass_oversamplers',
          'get_all_noisefilters',
          'generate_parameter_combinations']

def determine_n_quickest(oversamplers, n_quickest):
    """
    Determine the n quickest oversamplers.

    Args:
        oversamplers (list): a list of oversampler classes
        n_quickest (int): the number of quickest to be determined

    Returns:
        list: the n quickest oversamplers
    """

    if n_quickest == -1:
        return oversamplers

    pairs = [(osampler, runtimes[osampler.__name__])
                                for osampler in oversamplers]

    sorted_pairs = sorted(pairs, key=lambda x: x[1])

    n_quickest = np.min([n_quickest, len(oversamplers)])

    return [pair[0] for pair in sorted_pairs[:n_quickest]]


def get_all_oversamplers(n_quickest=-1):
    """
    Returns all oversampling classes

    Args:
        n_quickest (int): the number of quickest oversamplers to return

    Returns:
        list(OverSampling): list of all oversampling classes

    Example::

        import smote_variants as sv

        oversamplers= sv.get_all_oversamplers()
    """

    oversamplers = [A_SUWO, ADASYN, ADG, ADOMS, AHC, AMSCO, AND_SMOTE,
            ANS, ASMOBD, Assembled_SMOTE, Borderline_SMOTE1, Borderline_SMOTE2,
           CBSO, CCR, CE_SMOTE, cluster_SMOTE, CURE_SMOTE, DBSMOTE, DE_oversampling,
           DEAGO, distance_SMOTE, DSMOTE, DSRBF, E_SMOTE, Edge_Det_SMOTE, G_SMOTE,
           GASMOTE, Gaussian_SMOTE, polynom_fit_SMOTE_star, polynom_fit_SMOTE_bus,
           polynom_fit_SMOTE_poly, polynom_fit_SMOTE_mesh, Gazzah, IPADE_ID, ISMOTE,
           ISOMAP_Hybrid, KernelADASYN, kmeans_SMOTE, Lee, LLE_SMOTE, LN_SMOTE,
           LVQ_SMOTE, MCT, MDO, MOT2LD, MSYN, MWMOTE, NDO_sampling, NEATER,
           NRAS, NRSBoundary_SMOTE, NT_SMOTE, OUPS, PDFOS, ProWSyn, Random_SMOTE,
           ROSE, RWO_sampling, Safe_Level_SMOTE, SDSMOTE, Selected_SMOTE, SL_graph_SMOTE,
           SMMO, SMOBD, SMOTE_Cosine, SMOTE_D, SMOTE_ENN, SMOTE_FRST_2T, SMOTE_IPF,
           SMOTE_OUT, SMOTE_PSO, SMOTE_PSOBAT, SMOTE_RSB, SMOTE_TomekLinks, SMOTE,
           SN_SMOTE, SOI_CJ, SOMO, SPY, SSO, Stefanowski, SUNDO, Supervised_SMOTE,
           SVM_balance, SYMPROD, TRIM_SMOTE, V_SYNTH, VIS_RST, MSMOTE]

    return determine_n_quickest(oversamplers, n_quickest)

def get_metric_learning_oversamplers(n_quickest=-1):
    """
    Returns all oversampling classes supporting the use of metric learning
    for neighborhood calculations

    Args:
        n_quickest (int): the number of quickest oversamplers to return

    Returns:
        list(OverSampling): the list of all oversampling classes supporting
                            classifier induced distance

    """
    oversamplers = get_all_oversamplers()

    oversamplers = [o for o in oversamplers if
                    OverSampling.cat_metric_learning in o.categories]

    return determine_n_quickest(oversamplers, n_quickest)

def check_within_simplex_sampling(ss_params,
                                  within_simplex_sampling):
    """
    Check if the ss_params is conformant with the specification in
    within_simplex_sampling

    Args:
        ss_params (dict): the simplex sampling parameters
        within_simplex_sampling (str/list): the specification to check against

    Returns:
        bool: whether the specification is met
    """
    if isinstance(within_simplex_sampling, str):
        return ss_params['within_simplex_sampling'] == within_simplex_sampling

    return ss_params['within_simplex_sampling'] in within_simplex_sampling

def check_gaussian_component(ss_params,
                             with_gaussian_component):
    """
    Check if the ss_params is conformant with the specification of
    with_gaussian_component

    Args:
        ss_params (dict): the simplex sampling parameters
        with_gaussian_component (bool): the specification to check against

    Returns:
        bool: whether the specification is met
    """
    if with_gaussian_component is True:
        return (ss_params['gaussian_component'] is not None)\
                            and (len(ss_params['gaussian_component']) > 0)
    return not ((ss_params['gaussian_component'] is not None)\
                            and (len(ss_params['gaussian_component']) > 0))

def check_n_dim(ss_params,
                n_dim_range):
    """
    Check if the ss_params is conformant with the specification of
    n_dim_range

    Args:
        ss_params (dict): the simplex sampling parameters
        n_dim_range (int/list): the specification to check against

    Returns:
        bool: whether the specification is met
    """
    if isinstance(n_dim_range, int):
        return ss_params['n_dim'] == n_dim_range

    return ss_params['n_dim'] in n_dim_range

def get_simplex_sampling_oversamplers(within_simplex_sampling=None,
                                      exclude_within_simplex_sampling=None,
                                      with_gaussian_component=None,
                                      n_dim_range=None,
                                      n_quickest=-1):
    """
    Returns all oversampling classes supporting simplex sampling.

    Args:
        within_simplex_sampling (str/list/None): the value or values for the
                                            within simplex sampling strategies
        within_simplex_sampling (str/list/None): the value or values for the
                                            undesired within simplex sampling
                                            strategies
        with gaussian component (bool): only those techniques not having a
                                        Guassian component set by default
        n_dim_range (int/list): the desired n_dim value(s)
        n_quickest (int): the number of quickest oversamplers to return

    Returns:
        list(OverSampling): the list of all oversampling classes supporting
                            simplex sampling
    """

    oversamplers = get_all_oversamplers()

    results = []

    for os_class in oversamplers:
        if 'ss_params' not in list(inspect.signature(os_class).parameters.keys()):
            continue

        ss_params = None
        os_obj = os_class()

        if hasattr(os_obj, 'ss_params'):
            ss_params = os_obj.ss_params
        else:
            ss_params = SimplexSamplingMixin.get_params(os_obj)['ss_params']

        checks = []

        if within_simplex_sampling is not None:
            checks.append(check_within_simplex_sampling(ss_params,
                                                    within_simplex_sampling))

        if exclude_within_simplex_sampling is not None:
            checks.append(not check_within_simplex_sampling(ss_params,
                                            exclude_within_simplex_sampling))

        if with_gaussian_component is not None:
            checks.append(check_gaussian_component(ss_params,
                                                with_gaussian_component))

        if n_dim_range is not None:
            checks.append(check_n_dim(ss_params, n_dim_range))

        if len(checks) == 0 or all(checks):
            results.append(os_class)

    return determine_n_quickest(results, n_quickest)

def multiclass_filter(sampler):
    """
    Multiclass filter

    Args:
        sampler (obj): oversampler class

    Returns:
        bool: True if suitable for multiclass oversampling
    """
    term_0 = OverSampling.cat_changes_majority not in sampler.categories
    term_1 = OverSampling.cat_dim_reduction not in sampler.categories
    term_2 = 'proportion' in sampler().get_params()

    return term_0 and term_1 and term_2

def get_multiclass_oversamplers(strategy="eq_1_vs_many_successive",
                                    n_quickest=-1):
    """
    Returns all oversampling classes which can be used with the multiclass
    strategy specified

    Args:
        strategy (str): the multiclass oversampling strategy -
                        'eq_1_vs_many_successive'/'equalize_1_vs_many'
        n_quickest (int): the number of quickest oversamplers to return

    Returns:
        list(OverSampling): list of all oversampling classes which can be used
                            with the multiclass strategy specified

    Example::

        import smote_variants as sv

        oversamplers= sv.get_multiclass_oversamplers()

    """

    oversamplers = get_all_oversamplers()

    if strategy in ('eq_1_vs_many_successive', 'equalize_1_vs_many'):
        oversamplers = [o for o in oversamplers if multiclass_filter(o)]

    return determine_n_quickest(oversamplers, n_quickest)

def get_all_noisefilters():
    """
    Returns all noise filters
    Returns:
        list(NoiseFilter): list of all noise filter classes
    """
    return NoiseFilter.__subclasses__()


def generate_parameter_combinations(classes,
                                    n_max_comb=5,
                                    result_format='smote_variants',
                                    random_seed=5):
    """
    Return random parameter combinations for a set of oversamplers.

    Args:
        list(class): list of oversampling classes
        max_n_comb (int): the maximum number of parameter combinations
                            per oversampling technique
        result_format (str): 'dict'/'smote_variants' - the format of the results
        random_seed (int): the random seed to use
    """
    random_state = np.random.RandomState(random_seed)

    results = []

    for oversampler in classes:
        parameter_combinations = oversampler.parameter_combinations()
        random_state.shuffle(parameter_combinations)
        n_comb = np.min([len(parameter_combinations), n_max_comb])
        parameter_combinations = parameter_combinations[:n_comb]
        parameter_combinations = [oversampler(**pcomb).get_params()
                                    for pcomb in parameter_combinations]
        results.extend(parameter_combinations)

    if result_format == 'dict':
        return results

    return [('smote_variants', r['class_name'], r) for r in results]



"""
This module tests the OverSampling base class
"""

import numpy as np

import pytest

from smote_variants.base import (OverSampling, OverSamplingSimplex,
                            OverSamplingBase)

def test_oversamplingbase():
    """
    Testing the OverSampling base class
    """
    overs = OverSamplingBase()

    X = np.array([[1, 2, 3], [3, 4, 5], [6, 7, 8]])
    y = np.array([0, 1, 0])

    X_samp, _ = overs.sampling_algorithm(X, y)

    assert X_samp.shape == X.shape

    X_samp, _ = overs.sample(X, y)

    assert X_samp.shape == X.shape

    overs.checks={'min_n_min': 0}

    X_samp, _ = overs.sample(X, y)

    assert X_samp.shape == X.shape

    n_to_sample = overs.det_n_to_sample(1.0)

    assert n_to_sample >= 0

    n_to_sample = overs.det_n_to_sample(1.0, 20, 10)

    assert n_to_sample >= 0

    with pytest.raises(Exception):
        overs.det_n_to_sample('dummy')

    X_samp, _ = overs.fit_resample(X, y)

    assert X_samp.shape == X.shape

    X_samp, _ = overs.sample_with_timing(X, y)

    assert X_samp.shape == X.shape

    X_samp, _ = overs.return_copies(X, y, "message")

    assert X_samp.shape == X.shape

    X_prep = overs.preprocessing_transform(X) # pylint: disable=invalid-name

    assert X_prep.shape == X.shape

    assert len(overs.get_params()) == 0

    assert getattr(overs.set_params(**{'dummy': 1}), 'dummy') == 1

    assert len(overs.descriptor()) > 0

    assert len(str(overs)) > 0

def test_oversampling():
    """
    Testing the OverSampling class
    """
    overs = OverSampling(random_state=5)

    assert overs.get_params()['random_state'] == 5

    assert overs.sample_between_points(np.array([1]),
                                       np.array([2])).shape[0] == 1

    x_samp = overs.sample_between_points_componentwise(np.array([1]),
                                                       np.array([2]))

    assert x_samp.shape[0] == 1

    x_samp = overs.sample_between_points_componentwise(np.array([1]),
                                                       np.array([2]),
                                                       np.array([1]))

    assert x_samp.shape[0] == 1

    x_samp = overs.sample_by_jittering(np.array([1]), 1)

    assert x_samp.shape[0] == 1

    x_samp = overs.sample_by_jittering_componentwise(np.array([1]), 1)

    assert x_samp.shape[0] == 1

    x_samp = overs.sample_by_gaussian_jittering(np.array([1]), 1)

    assert x_samp.shape[0] == 1

def test_oversamplingsimplex():
    """
    Testing the OverSamplingSimplex class
    """
    with pytest.warns(Warning):
        overs = OverSamplingSimplex(random_state=5,
                                    checks={'simplex_dim': 3})

    assert overs.get_params()['random_state'] == 5



"""
This module implements the V_SYNTH method.
"""

import numpy as np

from sklearn.decomposition import PCA
from sklearn.metrics import pairwise_distances
import scipy.spatial as sspatial

from ..base import OverSampling
from .._logger import logger
_logger= logger

__all__= ['V_SYNTH']

class V_SYNTH(OverSampling):
    """
    References:
        * BibTex::

            @article{v_synth,
                     author = {Young,Ii, William A. and Nykl, Scott L. and
                                Weckman, Gary R. and Chelberg, David M.},
                     title = {Using Voronoi Diagrams to Improve
                                Classification Performances when Modeling
                                Imbalanced Datasets},
                     journal = {Neural Comput. Appl.},
                     issue_date = {July      2015},
                     volume = {26},
                     number = {5},
                     month = jul,
                     year = {2015},
                     issn = {0941-0643},
                     pages = {1041--1054},
                     numpages = {14},
                     url = {http://dx.doi.org/10.1007/s00521-014-1780-0},
                     doi = {10.1007/s00521-014-1780-0},
                     acmid = {2790665},
                     publisher = {Springer-Verlag},
                     address = {London, UK, UK},
                     keywords = {Data engineering, Data mining, Imbalanced
                                    datasets, Knowledge extraction,
                                    Numerical algorithms, Synthetic
                                    over-sampling},
                    }

    Notes:
        * The proposed encompassing bounding box generation is incorrect.
        * Voronoi diagram generation in high dimensional spaces is instable
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_sample_ordinary]

    def __init__(self,
                 proportion=1.0,
                 n_components=3,
                 *,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            n_components (int): number of components for PCA
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state, checks={'min_n_dim': 2})
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_components, "n_component", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_components = n_components
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_components': [3]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def add_bounding_box(self, X):
        """
        Add bounding box

        Args:
            X (np.array): all training vectors

        Returns:
            np.array, np.array: the extended set of vectors and the
                                bounding box
        """
        # creating the bounding box
        mins = np.min(X, axis=0)
        maxs = np.max(X, axis=0)
        mins = mins - 0.1*np.abs(mins)
        maxs = maxs + 0.1*np.abs(maxs)

        dim = len(X[0])

        def random_min_maxs():
            return np.where(self.random_state.randint(0, 1, size=dim) == 0,
                            mins,
                            maxs)

        n_bounding_box = min([100, len(X[0])])
        bounding_box = [random_min_maxs() for _ in range(n_bounding_box)]
        X_bb = np.vstack([X, bounding_box]) # pylint: disable=invalid-name

        return X_bb, bounding_box

    def filter_vectors_too_close(self,
                                X_pca, # pylint: disable=invalid-name
                                y_pca):
        """
        Determine the Voronoi-tessellation

        Args:
            X_pca (np.array): the vectors after PCA
            y_pca (np.array): the extended target labels

        Returns:
            np.array, np.array: the updated training vectors
                                and target labels
        """
        distm = pairwise_distances(X_pca)

        to_remove = []
        for idx in range(len(distm)):
            for jdx in range(idx+1, len(distm)):
                if distm[idx, jdx] < 0.001:
                    to_remove.append(idx)

        X_pca = np.delete(X_pca, to_remove, axis=0)
        y_pca = np.delete(y_pca, to_remove)

        return X_pca, y_pca

    def generate_samples(self, *,
                            voronoi,
                            y_pca,
                            candidate_face_generators,
                            n_to_sample):
        """
        Generate samples.

        Args:
            voronoi (obj): a Voronoi tessellation
            y_pca (np.array): the updated target labels
            candidate_face_generators (np.array): indices of the
                                                    face generators
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """
        samples = []
        for _ in range(n_to_sample):
            # randomly choosing a pair from the ridge point pairs of different
            # labels
            random_face = self.random_state.choice(candidate_face_generators)

            # extracting the vertices of the face between the points
            ridge_vertices = voronoi.ridge_vertices[random_face]
            face_vertices = voronoi.vertices[ridge_vertices]

            # creating a random vector for sampling the face (supposed to be
            # convex)
            weights = self.random_state.random_sample(size=len(face_vertices))
            weights = weights / np.sum(weights)

            # initiating a sample point on the face
            sample_point_on_face = np.sum((face_vertices.T * weights).T, axis=0)

            # finding the ridge point with the minority label
            if y_pca[voronoi.ridge_points[random_face][0]] == self.min_label:
                h_vec = voronoi.points[voronoi.ridge_points[random_face][0]]
            else:
                h_vec = voronoi.points[voronoi.ridge_points[random_face][1]]

            # generating a point between the minority ridge point and the
            # random point on the face
            samples.append(self.sample_between_points(sample_point_on_face,
                                                      h_vec))
        return np.vstack(samples)

    def dimensionality_reduction(self,
                                    X_bb, # pylint: disable=invalid-name
                                    y,
                                    bounding_box):
        """
        Do the dimensionality reduction.

        Args:
            X_bb (np.array): the extended set of vectors
            y (np.array): the target labels
            bounding_box (np.array): the bounding box vectors

        Returns:
            np.array, np.array, int, obj: the PCA-d training vectors, the
                                    corresponding target labels, the number
                                    of components and the fitted PCA object
        """
        # applying PCA to reduce the dimensionality of the data
        n_components = np.min([X_bb.shape[1], self.n_components])
        component_mask = np.array([True] * n_components)

        while np.sum(component_mask) > 1:
            pca = PCA(n_components=n_components)
            X_pca = pca.fit_transform(X_bb) # pylint: disable=invalid-name
            y_pca = np.hstack([y, np.repeat(-1, len(bounding_box))])

            component_mask = pca.explained_variance_ratio_ > 1e-2

            if np.sum(component_mask) == n_components:
                break

            n_components = np.sum(component_mask)

        return X_pca, y_pca, n_components, pca

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_bb, bounding_box = self.add_bounding_box(X) # pylint: disable=invalid-name

        X_pca, y_pca, n_components, pca = \
                        self.dimensionality_reduction(X_bb, y, bounding_box) # pylint: disable=invalid-name

        if n_components == 1:
            return self.return_copies(X, y,
                                "Voronoi tessellation in 1D is not feasible")

        X_pca, y_pca = self.filter_vectors_too_close(X_pca, y_pca) # pylint: disable=invalid-name

        # doing the Voronoi tessellation
        voronoi = sspatial.Voronoi(X_pca)

        # extracting those ridge point pairs which are candidates for
        # generating an edge between two cells of different class labels
        candidate_face_generators = []
        for idx, ridge in enumerate(voronoi.ridge_points):
            if ridge[0] < len(y) and ridge[1] < len(y) \
                            and y[ridge[0]] != y[ridge[1]]:
                candidate_face_generators.append(idx)

        # seems like this never happens
        if len(candidate_face_generators) == 0:
            return self.return_copies(X, y, "No candidate face generators found")

        # generating samples
        samples = self.generate_samples(voronoi=voronoi,
                                        y_pca=y_pca,
                                        candidate_face_generators=candidate_face_generators,
                                        n_to_sample=n_to_sample)

        return (np.vstack([X, pca.inverse_transform(samples)]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_components': self.n_components,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the CURE_SMOTE method.
"""

import numpy as np

from sklearn.preprocessing import MinMaxScaler

from ..base import coalesce_dict, coalesce
from ..base import pairwise_distances_mahalanobis

from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['CURE_SMOTE']

class CURE_SMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @Article{cure_smote,
                        author="Ma, Li
                        and Fan, Suohai",
                        title="CURE-SMOTE algorithm and hybrid algorithm for
                                feature selection and parameter optimization
                                based on random forests",
                        journal="BMC Bioinformatics",
                        year="2017",
                        month="Mar",
                        day="14",
                        volume="18",
                        number="1",
                        pages="169",
                        issn="1471-2105",
                        doi="10.1186/s12859-017-1578-z",
                        url="https://doi.org/10.1186/s12859-017-1578-z"
                        }

    Notes:
        * It is not specified how to determine the cluster with the
            "slowest growth rate"
        * All clusters can be removed as noise.
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_uses_clustering]

    def __init__(self,
                 proportion=1.0,
                 *,
                 n_clusters=5,
                 noise_th=2,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_clusters (int): number of clusters to generate
            noise_th (int): below this number of elements the cluster is
                                considered as noise
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_clusters, "n_clusters", 1)
        self.check_greater_or_equal(noise_th, "noise_th", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_clusters = n_clusters
        self.noise_th = noise_th
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_clusters': [5, 10, 15],
                                  'noise_th': [1, 3]}

        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def do_the_clustering(self, X_min, nn_params):
        """
        Do the clustering

        Args:
            X_min (np.array): the minority samples
            nn_params (dict): nearest neighbors (distance) parameters

        Returns:
            list(np.array): the clusters
        """
        # initiating clustering
        clusters = [np.array([i]) for i in range(len(X_min))]
        distm = pairwise_distances_mahalanobis(X_min,
                                        tensor=nn_params.get('metric_tensor', None))

        # setting the diagonal of the distance matrix to infinity
        for idx in range(len(distm)):
            distm[idx, idx] = np.inf

        # starting the clustering iteration
        iteration = 0
        while len(clusters) > self.n_clusters:
            iteration = iteration + 1

            # delete a cluster with slowest growth rate, determined by
            # the cluster size
            if iteration % self.n_clusters == 0:
                # extracting cluster sizes
                cluster_sizes = np.array([len(c) for c in clusters])
                # removing one of the clusters with the smallest size
                to_remove = np.where(cluster_sizes == np.min(cluster_sizes))[0]
                to_remove = self.random_state.choice(to_remove)
                del clusters[to_remove]
                # adjusting the distance matrix accordingly
                distm = np.delete(distm, to_remove, axis=0)
                distm = np.delete(distm, to_remove, axis=1)

            # finding the cluster pair with the smallest distance
            min_coord = np.where(distm == np.min(distm))
            merge_a = min_coord[0][0]
            merge_b = min_coord[1][0]

            # merging the clusters
            clusters[merge_a] = np.hstack(
                [clusters[merge_a], clusters[merge_b]])
            # removing one of them
            del clusters[merge_b]
            # adjusting the distances in the distance matrix
            distm[merge_a] = np.min(np.vstack([distm[merge_a], distm[merge_b]]), axis=0)
            distm[:, merge_a] = distm[merge_a]
            # removing the row and column corresponding to one of
            # the merged clusters
            distm = np.delete(distm, merge_b, axis=0)
            distm = np.delete(distm, merge_b, axis=1)
            # updating the diagonal
            for idx in range(len(distm)):
                distm[idx, idx] = np.inf

        # removing clusters declared as noise
        to_remove = []
        for idx, cluster in enumerate(clusters):
            if len(cluster) < self.noise_th:
                to_remove.append(idx)
        clusters = [clusters[i]
                    for i in range(len(clusters)) if i not in to_remove]

        return clusters

    def generate_samples_in_clusters(self, X_min, clusters, n_to_sample):
        """
        Generate samples within clusters

        Args:
            X_min (np.array): the minority samples
            clusters (list(np.array)): the cluster indices

        Returns:
            np.array: the generated samples
        """
        clusters_selected = self.random_state.choice(len(clusters), n_to_sample)
        cluster_unique, cluster_count = np.unique(clusters_selected,
                                                    return_counts=True)

        samples = []

        for idx, cluster in enumerate(cluster_unique):
            cluster_vectors = X_min[clusters[cluster]]
            cluster_mean = np.mean(cluster_vectors, axis=0)

            #if len(clusters[idx]) >= self.n_dim:
            X = np.array([cluster_mean])
            indices = np.array([np.hstack([np.array([0]),
                                np.arange(len(cluster_vectors))])])
            samples.append(self.sample_simplex(X=X,
                                                indices=indices,
                                                n_to_sample=cluster_count[idx],
                                                X_vertices=cluster_vectors))
            #else:
            #    sample_indices = self.random_state.choice(len(clusters[cluster]),
            #                                                cluster_count[idx])
            #    samples.append(cluster_vectors[sample_indices])

        return np.vstack(samples)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        # standardizing the data
        mms = MinMaxScaler()
        X_scaled = mms.fit_transform(X) # pylint: disable=invalid-name

        X_min = X_scaled[y == self.min_label]

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= \
                    self.metric_tensor_from_nn_params(nn_params, X, y)

        clusters = self.do_the_clustering(X_min, nn_params)

        # all clusters can be noise
        if len(clusters) == 0:
            return self.return_copies(X, y, "all clusters are removed as noise")

        samples = self.generate_samples_in_clusters(X_min,
                                                    clusters,
                                                    n_to_sample)

        return (np.vstack([X, mms.inverse_transform(np.vstack(samples))]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_clusters': self.n_clusters,
                'noise_th': self.noise_th,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the Lee method.
"""

import numpy as np

from ..base import coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['Lee']

class Lee(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @inproceedings{lee,
                             author = {Lee, Jaedong and Kim,
                                 Noo-ri and Lee, Jee-Hyong},
                             title = {An Over-sampling Technique with Rejection
                                        for Imbalanced Class Learning},
                             booktitle = {Proceedings of the 9th International
                                            Conference on Ubiquitous
                                            Information Management and
                                            Communication},
                             series = {IMCOM '15},
                             year = {2015},
                             isbn = {978-1-4503-3377-1},
                             location = {Bali, Indonesia},
                             pages = {102:1--102:6},
                             articleno = {102},
                             numpages = {6},
                             doi = {10.1145/2701126.2701181},
                             acmid = {2701181},
                             publisher = {ACM},
                             address = {New York, NY, USA},
                             keywords = {data distribution, data preprocessing,
                                            imbalanced problem, rejection rule,
                                            synthetic minority oversampling
                                            technique}
                            }
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_sample_ordinary,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 rejection_level=0.5,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors in nearest neighbor
                                component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            rejection_level (float): the rejection level of generated samples,
                                        if the fraction of majority labels in
                                        the local environment is higher than
                                        this number, the generated point is
                                        rejected
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        nn_params = coalesce(nn_params, {})

        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}

        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_in_range(rejection_level, "rejection_level", [0, 1])
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = nn_params
        self.rejection_level = rejection_level
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'rejection_level': [0.3, 0.5, 0.7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def generate_samples(self,
                            *,
                            y,
                            X_min,
                            n_to_sample,
                            ind_min,
                            nnmt):
        """
        Generate samples

        Args:
            y (np.array): all target labels
            X_min (np.array): minority samples
            n_to_sample (int): the number of samples to generate
            ind_min (np.array): the minority neighborhood structure
            nnmt (obj): a fitted nearest neighbor model

        Returns:
            np.array: the generated samples
        """
        samples = np.zeros(shape=(0, X_min.shape[1]), dtype=float)

        rejection_level = self.rejection_level
        tries = 0
        while samples.shape[0] < n_to_sample:
            # generating at least 20 samples always to avoid iterating
            # to generate 1 sample in each iteration
            n_missing = np.max([n_to_sample - samples.shape[0], 20])

            samples_new = self.sample_simplex(X=X_min,
                                                indices=ind_min,
                                                n_to_sample=n_missing)
            ind_new = nnmt.kneighbors(samples_new, return_distance=False)
            maj_frac = np.sum(y[ind_new] == self.maj_label, axis=1)\
                                                        /self.n_neighbors
            if np.sum(maj_frac < rejection_level) == 0:
                tries = tries + 1

            if tries > 5:
                rejection_level = rejection_level + 0.1
                tries = 0

            samples = np.vstack([samples,
                                 samples_new[maj_frac < rejection_level]])

        samples = samples[:n_to_sample]

        return samples

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not neeed")

        X_min = X[y == self.min_label]

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        # fitting nearest neighbors models to find neighbors of minority
        # samples in the total data and in the minority datasets
        n_neighbors = min([len(X_min), self.n_neighbors])
        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nnmt.fit(X)

        n_neighbors = min([len(X_min), self.n_neighbors + 1])
        nn_min= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                    n_jobs=self.n_jobs,
                                                    **nn_params)
        nn_min.fit(X_min)
        ind_min = nn_min.kneighbors(X_min, return_distance=False)

        # do the sampling, we implemented a continuous tweaking of rejection
        # levels in order to fix situations when no unrejectable data can
        # be can be generated

        samples = self.generate_samples(y=y,
                                        X_min=X_min,
                                        n_to_sample=n_to_sample,
                                        ind_min=ind_min,
                                        nnmt=nnmt)

        #samples = []
        #passed = 0
        #trial = 0
        #rejection_level = self.rejection_level
        #while len(samples) < n_to_sample:
        #    # checking if we managed to generate a single data in 1000 trials
        #    if passed == trial and passed > 1000:
        #        rejection_level = rejection_level + 0.1
        #        trial = 0
        #        passed = 0
        #    trial = trial + 1
        #    # generating random point
        #    idx = self.random_state.randint(len(X_min))
        #    random_neighbor_idx = self.random_state.choice(ind_min[idx][1:])
        #    X_a = X_min[idx]
        #    X_b = X_min[random_neighbor_idx]
        #    random_point = self.sample_between_points(X_a, X_b)
        #    # checking if the local environment is above the rejection level
        #    ind_new = nnmt.kneighbors(random_point.reshape(1, -1), return_distance=False)
        #    maj_frac = np.sum(y[ind_new[0]] == self.maj_label)/self.n_neighbors
        #
        #    if maj_frac < rejection_level:
        #        samples.append(random_point)
        #    else:
        #        passed = passed + 1

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'rejection_level': self.rejection_level,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
A module for global package configurations.
"""

__all__=['suppress_external_warnings',
         'suppress_internal_warnings',
         'DEFAULT_SUPPRESS_EXTERNAL_WARNINGS',
         'DEFAULT_SUPPRESS_INTERNAL_WARNINGS']

DEFAULT_SUPPRESS_EXTERNAL_WARNINGS = True
DEFAULT_SUPPRESS_INTERNAL_WARNINGS = True

flags = {'SUPPRESS_EXTERNAL_WARNINGS': DEFAULT_SUPPRESS_EXTERNAL_WARNINGS,
         'SUPPRESS_INTERNAL_WARNINGS': DEFAULT_SUPPRESS_INTERNAL_WARNINGS}

def suppress_external_warnings(flag=None):
    """
    Set/get the value of SUPPRESS_EXTERNAL_WARNINGS

    Args:
        flag (bool): the new flag

    Returns:
        bool: the current value of the flag
    """

    if flag is not None:
        flags['SUPPRESS_EXTERNAL_WARNINGS'] = flag
    return flags['SUPPRESS_EXTERNAL_WARNINGS']

def suppress_internal_warnings(flag=None):
    """
    Set/get the value of SUPPRESS_INTERNAL_WARNINGS

    Args:
        flag (bool): the new flag

    Returns:
        bool: the current value of the flag
    """

    if flag is not None:
        flags['SUPPRESS_INTERNAL_WARNINGS'] = flag
    return flags['SUPPRESS_INTERNAL_WARNINGS']



"""
This module implements the NT_SMOTE method.
"""

import numpy as np

from ..base import coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['NT_SMOTE']

class NT_SMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{nt_smote,
                            author={Xu, Y. H. and Li, H. and Le, L. P. and
                                        Tian, X. Y.},
                            booktitle={2014 Seventh International Joint
                                        Conference on Computational Sciences
                                        and Optimization},
                            title={Neighborhood Triangular Synthetic Minority
                                    Over-sampling Technique for Imbalanced
                                    Prediction on Small Samples of Chinese
                                    Tourism and Hospitality Firms},
                            year={2014},
                            volume={},
                            number={},
                            pages={534-538},
                            keywords={financial management;pattern
                                        classification;risk management;sampling
                                        methods;travel industry;Chinese
                                        tourism; hospitality firms;imbalanced
                                        risk prediction;minority class samples;
                                        up-sampling approach;neighborhood
                                        triangular synthetic minority
                                        over-sampling technique;NT-SMOTE;
                                        nearest neighbor idea;triangular area
                                        sampling idea;single classifiers;data
                                        excavation principles;hospitality
                                        industry;missing financial indicators;
                                        financial data filtering;financial risk
                                        prediction;MDA;DT;LSVM;logit;probit;
                                        firm risk prediction;Joints;
                                        Optimization;imbalanced datasets;
                                        NT-SMOTE;neighborhood triangular;
                                        random sampling},
                            doi={10.1109/CSO.2014.104},
                            ISSN={},
                            month={July}}
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_application,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 nn_params=None,
                 ss_params=None,
                 *,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 3, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        # find two nearest minority samples
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=self.n_dim,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_min)
        ind = nnmt.kneighbors(X_min, return_distance=False)

        samples = self.sample_simplex(X=X_min,
                                        indices=ind,
                                        n_to_sample=n_to_sample)

        #samples = []
        #while len(samples) < n_to_sample:
        #    # select point randomly
        #    idx = self.random_state.randint(len(X_min))
        #    P_1 = X_min[idx]
        #    # find two closest neighbors
        #    P_2 = X_min[ind[idx][1]]
        #    P_3 = X_min[ind[idx][2]]
        #    # generate random point by sampling the specified triangle
        #    r_1 = self.random_state.random_sample()
        #    r_2 = self.random_state.random_sample()
        #    samples.append((P_3 + r_1 * ((P_1 + r_2 * (P_2 - P_1)) - P_3)))

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
Testing the E_SMOTE.
"""

import numpy as np

from smote_variants import E_SMOTE

np.random.seed(42)
X = np.random.normal(size=(1000, 30))
y = np.hstack([np.repeat(1, 70), np.repeat(0, 930)])

def test_specific():
    """
    Oversampler specific testing
    """

    obj = E_SMOTE()

    X_samp, _ = obj.sample(X, y)

    assert len(X_samp) > 0
